/*
								+---------------------------------+
								|                                 |
								|   ***   OGL state guard   ***   |
								|                                 |
								|  Copyright   -tHE SWINe- 2005  |
								|                                 |
								|         OpenGLState.cpp         |
								|                                 |
								+---------------------------------+
*/

/*
 *	passed code revision at 05-16-2006
 *
 *  integer used to store just true or false were changed to bool
 *	fixed some minor errors with stencil write-mask
 *  body of CGLState::Update_States moved to TGLState::Update where it belongs
 *
 *	07-15-2006
 *
 *	modified PolygonOffset support to match OpenGL 1.1 (EXT is actualy older)
 *
 *	07-25-2006
 *
 *	Files GlState.cpp and GlState.h renamed to OpenGLState.cpp and OpenGLState.h
 *
 */

#include "CallStack.h"
#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <math.h>
#include <gl/gl.h>
#include <vector>

#include "glext.h"
#include "Wglext.h"
//#include "OpenGLDrv.h"
#include "OpenGL20.h"
#include "OpenGLstate.h"

#if defined(_MSC_VER) && !defined(__MWERKS__) && !defined(for)
#define for if(0) {} else for
#endif

/*
 *								=== TGLState ===
 */

/*
 *	void TGLState::Update()
 *		- read current gl-state into itself
 *		- note it may take a few milliseconds, not a candidate to be called every frame
 */
void TGLState::Update()
{
	__FuncGuard("TGLState::Update");

	glGetFloatv(GL_CURRENT_COLOR, /*this->*/f_color);
	for(int i = 0; i < 4; i ++)
		/*this->*/n_color[i] = (unsigned char)(/*this->*/f_color[i] * 255.0f);
	// color

	/*this->*/b_blend = 0 != glIsEnabled(GL_BLEND);
	glGetIntegerv(GL_BLEND_SRC, (int*)&/*this->*/n_blend_func[0]);
	glGetIntegerv(GL_BLEND_DST, (int*)&/*this->*/n_blend_func[1]);
	// blending

	glGetIntegerv(GL_FRONT_FACE, (int*)&/*this->*/n_front_face);
	/*this->*/b_cull_face = 0 != glIsEnabled(GL_CULL_FACE);
	glGetIntegerv(GL_CULL_FACE_MODE, (int*)&/*this->*/n_cull_face);
	// face culling

	GLboolean b_tmp_quad[4];
	glGetBooleanv(GL_DEPTH_WRITEMASK, b_tmp_quad);
	/*this->*/b_depth_write = 0 != b_tmp_quad[0];
	glGetBooleanv(GL_COLOR_WRITEMASK, b_tmp_quad);
	/*this->*/b_color_write[0] = 0 != b_tmp_quad[0];
	/*this->*/b_color_write[1] = 0 != b_tmp_quad[1];
	/*this->*/b_color_write[2] = 0 != b_tmp_quad[2];
	/*this->*/b_color_write[3] = 0 != b_tmp_quad[3];
	glGetIntegerv(GL_STENCIL_WRITEMASK, (int*)&/*this->*/n_stencil_write[1]); // integer - bit mask really
	// write masks (updated - bool might not be the size of GLboolean)

	glGetFloatv(GL_COLOR_CLEAR_VALUE, /*this->*/f_clr_color);
	glGetDoublev(GL_DEPTH_CLEAR_VALUE, &/*this->*/f_clr_depth);
	glGetIntegerv(GL_STENCIL_CLEAR_VALUE, &/*this->*/n_clr_stencil);
	// clear color

	/*this->*/b_alpha_test = 0 != glIsEnabled(GL_ALPHA_TEST);
	glGetIntegerv(GL_ALPHA_TEST_FUNC, (int*)&/*this->*/n_alpha_func);
	glGetFloatv(GL_ALPHA_TEST_REF, &/*this->*/f_alpha_tresh);
	// alpha test

	/*this->*/b_depth_test = 0 != glIsEnabled(GL_DEPTH_TEST);
	glGetIntegerv(GL_DEPTH_FUNC, (int*)&/*this->*/n_depth_func);
	// depth test

	CGLExtensionHandler::n_GetGL20FuncPointers();
	/*this->*/b_separate_gl2_stencil =
		CGLExtensionHandler::b_Support_OpenGL(2, 0) &&
		glStencilFuncSeparate && glStencilOpSeparate && glStencilMaskSeparate;
	//
	/*this->*/b_stencil_test = 0 != glIsEnabled(GL_STENCIL_TEST);
	glGetIntegerv(GL_STENCIL_FUNC, (int*)&/*this->*/n_stencil_func[1]);
	glGetIntegerv(GL_STENCIL_REF, &/*this->*/n_stencil_ref[1]);
	glGetIntegerv(GL_STENCIL_VALUE_MASK, (int*)&/*this->*/n_stencil_mask[1]);
	glGetIntegerv(GL_STENCIL_FAIL, (int*)&/*this->*/n_stencil_op_sfail[1]);
	glGetIntegerv(GL_STENCIL_PASS_DEPTH_FAIL, (int*)&/*this->*/n_stencil_op_dfail[1]);
	glGetIntegerv(GL_STENCIL_PASS_DEPTH_PASS, (int*)&/*this->*/n_stencil_op_dpass[1]);
	if(/*this->*/b_separate_gl2_stencil) {
		glGetIntegerv(GL_STENCIL_BACK_FUNC, (int*)&/*this->*/n_stencil_func[0]);
		glGetIntegerv(GL_STENCIL_BACK_REF, &/*this->*/n_stencil_ref[0]);
		glGetIntegerv(GL_STENCIL_BACK_VALUE_MASK, (int*)&/*this->*/n_stencil_mask[0]);
		glGetIntegerv(GL_STENCIL_BACK_FAIL, (int*)&/*this->*/n_stencil_op_sfail[0]);
		glGetIntegerv(GL_STENCIL_BACK_PASS_DEPTH_FAIL,
			(int*)&/*this->*/n_stencil_op_dfail[0]);
		glGetIntegerv(GL_STENCIL_BACK_PASS_DEPTH_PASS,
			(int*)&/*this->*/n_stencil_op_dpass[0]);
		glGetIntegerv(GL_STENCIL_BACK_WRITEMASK, (int*)&/*this->*/n_stencil_write[0]);
	} else {
		/*this->*/n_stencil_func[0] = /*this->*/n_stencil_func[1];
		/*this->*/n_stencil_ref[0] = /*this->*/n_stencil_ref[1];
		/*this->*/n_stencil_mask[0] = /*this->*/n_stencil_mask[1];
		/*this->*/n_stencil_op_sfail[0] = /*this->*/n_stencil_op_sfail[1];
		/*this->*/n_stencil_op_dfail[0] = /*this->*/n_stencil_op_dfail[1];
		/*this->*/n_stencil_op_dpass[0] = /*this->*/n_stencil_op_dpass[1];
		/*this->*/n_stencil_write[0] = /*this->*/n_stencil_write[1];
	}
	// stencil test

#ifdef __MULTITEXTURE_ENABLED
	if(CGLExtensionHandler::n_GetMultitextureFuncPointers()) {
		_ASSERTE(0);
	}

	glGetIntegerv(GL_MAX_TEXTURE_UNITS, &/*this->*/n_max_texture_units);
	glGetIntegerv(GL_ACTIVE_TEXTURE, &/*this->*/n_active_tex_unit);
	/*this->*/n_active_tex_unit -= GL_TEXTURE0;
	_ASSERTE(/*this->*/n_max_texture_units <= 32);
	//
	for(int i = 0; i < /*this->*/n_max_texture_units; i ++) {
		glActiveTexture(GL_TEXTURE0 + i);
		//
		/*this->*/b_texture1d_enabled[i] = 0 != glIsEnabled(GL_TEXTURE_1D);
		/*this->*/b_texture2d_enabled[i] = 0 != glIsEnabled(GL_TEXTURE_2D);
		/*this->*/b_texture3d_enabled[i] = false;
		/*this->*/n_texture_cubemap[i] = false; // nkterej hw tohle nemus podporovat
		/*this->*/b_texture3d_enabled[i] = 0 != glIsEnabled(GL_TEXTURE_3D);
		/*this->*/b_texture_cubemap_enabled[i] = 0 != glIsEnabled(GL_TEXTURE_CUBE_MAP);
		//
		glGetIntegerv(GL_TEXTURE_1D, (int*)&/*this->*/n_texture1d[i]);
		glGetIntegerv(GL_TEXTURE_2D, (int*)&/*this->*/n_texture2d[i]);
		glGetIntegerv(GL_TEXTURE_3D, (int*)&/*this->*/n_texture3d[i]);
		glGetIntegerv(GL_TEXTURE_CUBE_MAP, (int*)&/*this->*/n_texture_cubemap[i]);
	}
	glActiveTexture(/*this->*/n_active_tex_unit + GL_TEXTURE0);
	// texture units
#else
	/*this->*/n_max_texture_units = 1;
	/*this->*/n_active_tex_unit = 0;
	//
	/*this->*/b_texture1d_enabled[0] = 0 != glIsEnabled(GL_TEXTURE_1D);
	/*this->*/b_texture2d_enabled[0] = 0 != glIsEnabled(GL_TEXTURE_2D);
	/*this->*/b_texture3d_enabled[0] = false;
	/*this->*/n_texture_cubemap[0] = false; // nkterej hw tohle nemus podporovat
	/*this->*/b_texture3d_enabled[0] = 0 != glIsEnabled(GL_TEXTURE_3D);
	/*this->*/n_texture_cubemap[0] = 0 != glIsEnabled(GL_TEXTURE_CUBE_MAP);
	//
	glGetIntegerv(GL_TEXTURE_1D, &/*this->*/n_texture1d[0]);
	glGetIntegerv(GL_TEXTURE_2D, &/*this->*/n_texture2d[0]);
	glGetIntegerv(GL_TEXTURE_3D, &/*this->*/n_texture3d[0]);
	glGetIntegerv(GL_TEXTURE_CUBE_MAP, &/*this->*/n_texture_cubemap[0]);
	// texture units
#endif

	/*this->*/b_lighting = 0 != glIsEnabled(GL_LIGHTING);
	glGetIntegerv(GL_MAX_LIGHTS, &/*this->*/n_max_lights);
	_ASSERTE(/*this->*/n_max_lights <=32);
	for(int i = 0; i < /*this->*/n_max_lights; i ++) {
		/*this->*/b_light[i] = 0 != glIsEnabled(GL_LIGHT0 + i);
		//
		glGetLightfv(GL_LIGHT0 + i, GL_AMBIENT, /*this->*/t_light_data[i].f_ambient);
		glGetLightfv(GL_LIGHT0 + i, GL_DIFFUSE, /*this->*/t_light_data[i].f_diffuse);
		glGetLightfv(GL_LIGHT0 + i, GL_SPECULAR, /*this->*/t_light_data[i].f_specular);
		glGetLightfv(GL_LIGHT0 + i, GL_POSITION, /*this->*/t_light_data[i].f_position);
		glGetLightfv(GL_LIGHT0 + i, GL_SPOT_DIRECTION,
			/*this->*/t_light_data[i].f_spot_direction);
		glGetLightfv(GL_LIGHT0 + i, GL_SPOT_EXPONENT,
			&/*this->*/t_light_data[i].f_spot_exponent);
		glGetLightfv(GL_LIGHT0 + i, GL_SPOT_CUTOFF,
			&/*this->*/t_light_data[i].f_spot_cutoff);
		glGetLightfv(GL_LIGHT0 + i, GL_CONSTANT_ATTENUATION,
			&/*this->*/t_light_data[i].f_constant_attenuation);
		glGetLightfv(GL_LIGHT0 + i, GL_LINEAR_ATTENUATION,
			&/*this->*/t_light_data[i].f_linear_attenuation);
		glGetLightfv(GL_LIGHT0 + i, GL_QUADRATIC_ATTENUATION,
			&/*this->*/t_light_data[i].f_quadratic_attenuation);
	}
	// lights

	glGetFloatv(GL_LIGHT_MODEL_AMBIENT, /*this->*/f_model_ambient);
	glGetBooleanv(GL_LIGHT_MODEL_LOCAL_VIEWER, &/*this->*/b_model_local_viewer);
	glGetBooleanv(GL_LIGHT_MODEL_TWO_SIDE, &/*this->*/b_model_two_side);
	glGetIntegerv(GL_LIGHT_MODEL_COLOR_CONTROL, (int*)&/*this->*/n_model_color_control);
	// lighting model

	for(int i = 0; i < 2; i ++) {
		glGetMaterialfv((i)? GL_FRONT : GL_BACK, GL_AMBIENT,
			/*this->*/f_material_ambient[i]);
		glGetMaterialfv((i)? GL_FRONT : GL_BACK, GL_DIFFUSE,
			/*this->*/f_material_diffuse[i]);
		glGetMaterialfv((i)? GL_FRONT : GL_BACK, GL_SPECULAR,
			/*this->*/f_material_specular[i]);
		glGetMaterialfv((i)? GL_FRONT : GL_BACK, GL_EMISSION,
			/*this->*/f_material_emission[i]);
		glGetMaterialfv((i)? GL_FRONT : GL_BACK, GL_SHININESS,
			&/*this->*/f_material_shininess[i]);
	}
	// material

	/*this->*/b_fog = 0 != glIsEnabled(GL_FOG);
	glGetIntegerv(GL_FOG_MODE, (int*)&/*this->*/n_fog_type);
	glGetFloatv(GL_FOG_COLOR, /*this->*/f_fog_color);
	glGetFloatv(GL_FOG_START, &/*this->*/f_fog_start);
	glGetFloatv(GL_FOG_DENSITY, &/*this->*/f_fog_density);
	glGetFloatv(GL_FOG_END, &/*this->*/f_fog_end);
	// fog

	glGetIntegerv(GL_SHADE_MODEL, (int*)&/*this->*/n_shade_model);
	// shade model

	glGetIntegerv(GL_POLYGON_MODE, (int*)/*this->*/n_polygon_mode);
	glGetFloatv(GL_LINE_WIDTH, &/*this->*/f_line_width);
	glGetFloatv(GL_POINT_SIZE, &/*this->*/f_point_size);
	// rasterizer props

	for(int j = 0; j < /*this->*/n_max_texture_units; j ++) {
		glActiveTexture(j + GL_TEXTURE0);
		for(int i = 0; i < 4; i ++) {
			/*this->*/b_texgen_enabled[j][i] = 0 != glIsEnabled(GL_TEXTURE_GEN_S + i);
			//
			glGetTexGeniv(GL_S + i, GL_TEXTURE_GEN_MODE, (int*)&/*this->*/n_texgen_mode[j][i]);
			glGetTexGenfv(GL_S + i, GL_EYE_PLANE, /*this->*/f_texgen_eye_plane[j][i]);
			glGetTexGenfv(GL_S + i, GL_OBJECT_PLANE, /*this->*/f_texgen_object_plane[j][i]);
		}
	}
	glActiveTexture(/*this->*/n_active_tex_unit + GL_TEXTURE0);
	// texgen props

	/*this->*/b_polygon_offset_support = CGLExtensionHandler::b_Support_OpenGL(1, 1);/*
		CGLExtensionHandler::b_SupportedExtension("GL_ARB_polygon_offset") &&
		!CGLExtensionHandler::n_GetPolygonOffsetARBFuncPointers();*/
	//
	if(/*this->*/b_polygon_offset_support) {
		/*this->*/b_polygon_offset[0] = 0 != glIsEnabled(GL_POLYGON_OFFSET_FILL);
		/*this->*/b_polygon_offset[1] = 0 != glIsEnabled(GL_POLYGON_OFFSET_LINE);
		/*this->*/b_polygon_offset[2] = 0 != glIsEnabled(GL_POLYGON_OFFSET_POINT);
		f_polygon_offset_factor = 0;
		f_polygon_offset_bias = 0; // what to do?
		/*glGetFloatv(GL_POLYGON_OFFSET_FACTOR_EXT, &/*this->* /f_polygon_offset_factor);
		glGetFloatv(GL_POLYGON_OFFSET_BIAS_EXT, &/*this->* /f_polygon_offset_bias);*/
	} else {
		/*this->*/b_polygon_offset[0] = false;
		/*this->*/b_polygon_offset[1] = false;
		/*this->*/b_polygon_offset[2] = false;
	}

	/*this->*/b_fragment_program_support =
		CGLExtensionHandler::b_SupportedExtension("GL_ARB_fragment_program") &&
		!CGLExtensionHandler::n_GetFragmentProgramARBFuncPointers();
	/*this->*/b_vertex_program_support =
		CGLExtensionHandler::b_SupportedExtension("GL_ARB_vertex_program") &&
		!CGLExtensionHandler::n_GetVertexProgramARBFuncPointers();
	//
	if(/*this->*/b_vertex_program_support) {
		/*this->*/b_fragment_program_enabled = 0 != glIsEnabled(GL_VERTEX_PROGRAM_ARB);
		glGetIntegerv(GL_VERTEX_PROGRAM_ARB, (int*)&/*this->*/n_vertex_program);
	} else {
		/*this->*/n_vertex_program = 0;
		/*this->*/b_vertex_program_enabled = false;
	}
	if(/*this->*/b_fragment_program_support) {
		/*this->*/b_fragment_program_enabled = 0 != glIsEnabled(GL_FRAGMENT_PROGRAM_ARB);
		glGetIntegerv(GL_FRAGMENT_PROGRAM_ARB, (int*)&/*this->*/n_fragment_program);
	} else {
		/*this->*/n_fragment_program = 0;
		/*this->*/b_fragment_program_enabled = false;
	}
	// low-level shading programmability

	/*this->*/b_shader_support =
		(CGLExtensionHandler::b_SupportedExtension("GL_ARB_fragment_shader") ||
		(CGLExtensionHandler::b_SupportedExtension("GL_ARB_vertex_shader") &&
		!CGLExtensionHandler::n_GetVertexShaderARBFuncPointers())) &&
		CGLExtensionHandler::b_SupportedExtension("GL_ARB_shader_objects") &&
		CGLExtensionHandler::b_SupportedExtension("GL_ARB_shading_language_100") &&
		!CGLExtensionHandler::n_GetShaderObjectARBFuncPointers();
	/*this->*/b_shader_gl2_support =
		CGLExtensionHandler::b_Support_OpenGL(2, 0) &&
		(CGLExtensionHandler::b_SupportedExtension("GL_ARB_fragment_shader") ||
		CGLExtensionHandler::b_SupportedExtension("GL_ARB_vertex_shader")) &&
		CGLExtensionHandler::b_SupportedExtension("GL_ARB_shading_language_100") &&
		!CGLExtensionHandler::n_GetGL20FuncPointers();
	//
	if(/*this->*/b_shader_gl2_support) {
		glGetIntegerv(GL_CURRENT_PROGRAM, (int*)&/*this->*/n_program_object);
	} else if(/*this->*/b_shader_support) {
		/*this->*/n_program_object = glGetHandleARB(GL_PROGRAM_OBJECT_ARB);
	} else
		/*this->*/n_program_object = 0;
	// high-level ARB shading programmability
	// high-level GL20 shading programmability

	glGetIntegerv(GL_MAX_CLIP_PLANES, &n_max_clip_planes);
	for(int i = 0; i < n_max_clip_planes; i ++) {
		b_clip_plane[i] = 0 != glIsEnabled(GL_CLIP_PLANE0 + i);
		glGetClipPlane(GL_CLIP_PLANE0 + i, f_clip_plane[i]);
	}
	// clip planes
}

/*
 *								=== ~TGLState ===
 */

/*
 *								=== CGLState ===
 */

std::vector<TGLState*> *CGLState::m_p_state = 0;
int CGLState::m_n_state_refs = 0;

CGLState::CGLState(HDC h_dc, HGLRC h_glrc)
	:m_h_dc(0), m_h_glrc(0), m_b_init(false)
{
	__FuncGuard("CGLState::CGLState");

	TGLState *p_my_state;
	HGLRC h_prev_glrc;
	HDC h_prev_dc;

	if(!m_p_state) {
		if(!(m_p_state = new std::vector<TGLState*>))
			return;
	}
	m_n_state_refs ++;
	// first class

	m_h_dc = h_dc;
	m_h_glrc = h_glrc;

	bool b_found = false;
	for(int i = 0; i < m_p_state->size(); i ++) {
		if((*m_p_state)[i]->h_glrc == h_glrc) {
			m_p_cur_state = (*m_p_state)[i];
			m_p_cur_state->n_references ++;

			b_found = true;
			break;
		}
	}
	if(!b_found) {
		if(!(p_my_state = new TGLState))
			return;

		p_my_state->h_glrc = m_h_glrc;
		p_my_state->n_references = 1;
		//
		m_p_state->push_back(p_my_state);
		_ASSERTE(m_p_state->size());
		m_p_cur_state = (*m_p_state)[m_p_state->size() - 1];
		//
		h_prev_dc = wglGetCurrentDC();
		h_prev_glrc = wglGetCurrentContext();
		wglMakeCurrent(m_h_dc, m_h_glrc);
		//
		Update_States();
		//
		wglMakeCurrent(h_prev_dc, h_prev_glrc);
		// create my state and store current opengl state isnide
	}
	// try to find state that was already set for this context

	m_b_init = true;
}

CGLState::~CGLState()
{
	__FuncGuard("CGLState::~CGLState");

	if(m_b_init && !(-- m_n_state_refs) && m_p_state) {
		for(int i = 0; i < m_p_state->size(); i ++)
			delete (*m_p_state)[i];
		delete m_p_state;
		m_p_state = 0;
	}
	m_b_init = false;
}

void CGLState::Update_States()
{
	__FuncGuard("CGLState::Update_States");

	_ASSERTE(m_p_cur_state);

	m_p_cur_state->Update();

	_ASSERTE(glGetError() == GL_NO_ERROR);
}

HDC CGLState::h_DeviceContext()
{
	__FuncGuard("CGLState::h_DeviceContext");

	return m_h_dc;
}

HGLRC CGLState::h_RenderingContext()
{
	__FuncGuard("CGLState::h_RenderingContext");

	return m_h_glrc;
}

void __fastcall CGLState::LineWidth(GLfloat f_width)
{
	__FuncGuard("CGLState::LineWidth");

	if(m_p_cur_state->f_line_width != f_width)
		glLineWidth(m_p_cur_state->f_line_width = f_width);
}

void __fastcall CGLState::PointSize(GLfloat f_size)
{
	__FuncGuard("CGLState::PointSize");

	if(m_p_cur_state->f_point_size != f_size)
		glPointSize(m_p_cur_state->f_point_size = f_size);
}

void __fastcall CGLState::PolygonMode(GLenum n_side, GLenum n_mode)
{
	__FuncGuard("CGLState::PolygonMode");

	if(n_side == GL_FRONT_AND_BACK &&
	   (n_mode != m_p_cur_state->n_polygon_mode[0] ||
	   n_mode != m_p_cur_state->n_polygon_mode[1])) {
		glPolygonMode(GL_FRONT_AND_BACK, m_p_cur_state->n_polygon_mode[0] =
			m_p_cur_state->n_polygon_mode[1] = n_mode);
	} else if(n_side == GL_FRONT &&
	   n_mode != m_p_cur_state->n_polygon_mode[0])
		glPolygonMode(GL_FRONT, m_p_cur_state->n_polygon_mode[0] = n_mode);
	else if(n_side == GL_BACK &&
	   n_mode != m_p_cur_state->n_polygon_mode[1])
		glPolygonMode(GL_BACK, m_p_cur_state->n_polygon_mode[1] = n_mode);	
}

void __fastcall CGLState::EnableFog()
{
	__FuncGuard("CGLState::EnableFog");

	if(!m_p_cur_state->b_fog) {
		m_p_cur_state->b_fog = true;
		glEnable(GL_FOG);
	}
}

void __fastcall CGLState::DisableFog()
{
	__FuncGuard("CGLState::DisableFog");

	if(m_p_cur_state->b_fog) {
		m_p_cur_state->b_fog = false;
		glDisable(GL_FOG);
	}
}

void __fastcall CGLState::FogMode(GLenum n_value)
{
	__FuncGuard("CGLState::FogMode");

	if(m_p_cur_state->n_fog_type != n_value)
		glFogi(GL_FOG_MODE, m_p_cur_state->n_fog_type = n_value);
}

void __fastcall CGLState::FogDensity(GLfloat f_value)
{
	__FuncGuard("CGLState::FogDensity");

	if(m_p_cur_state->f_fog_density != f_value)
		glFogf(GL_FOG_DENSITY, m_p_cur_state->f_fog_density = f_value);
}

void __fastcall CGLState::FogColor(const GLfloat *p_color)
{
	__FuncGuard("CGLState::FogColor");

	if(memcmp(m_p_cur_state->f_fog_color, p_color, 4 * sizeof(GLfloat))) {
		memcpy(m_p_cur_state->f_fog_color, p_color, 4 * sizeof(GLfloat));
		glFogfv(GL_FOG_COLOR, p_color);
	}
}

void __fastcall CGLState::FogStart(GLfloat f_value)
{
	__FuncGuard("CGLState::FogStart");

	if(m_p_cur_state->f_fog_start != f_value)
		glFogf(GL_FOG_START, m_p_cur_state->f_fog_start = f_value);
}

void __fastcall CGLState::FogEnd(GLfloat f_value)
{
	__FuncGuard("CGLState::FogEnd");

	if(m_p_cur_state->f_fog_end != f_value)
		glFogf(GL_FOG_END, m_p_cur_state->f_fog_end = f_value);
}

void __fastcall CGLState::EnableLighting()
{
	__FuncGuard("CGLState::EnableLighting");

	if(!m_p_cur_state->b_lighting) {
		glEnable(GL_LIGHTING);
		m_p_cur_state->b_lighting = true;
	}
}

void __fastcall CGLState::DisableLighting()
{
	__FuncGuard("CGLState::DisableLighting");

	if(m_p_cur_state->b_lighting) {
		glDisable(GL_LIGHTING);
		m_p_cur_state->b_lighting = false;
	}
}

void __fastcall CGLState::EnableLight(int n_light)
{
	__FuncGuard("CGLState::EnableLight");

	_ASSERTE(n_light >= 0 && n_light < m_p_cur_state->n_max_lights);

	if(!m_p_cur_state->b_light[n_light]) {
		glEnable(GL_LIGHT0 + n_light);
		m_p_cur_state->b_light[n_light] = true;
	}
}

void __fastcall CGLState::DisableLight(int n_light)
{
	__FuncGuard("CGLState::DisableLight");

	_ASSERTE(n_light >= 0 && n_light < m_p_cur_state->n_max_lights);

	if(m_p_cur_state->b_light[n_light]) {
		glDisable(GL_LIGHT0 + n_light);
		m_p_cur_state->b_light[n_light] = false;
	}
}

void __fastcall CGLState::ShadeModel(GLenum n_model)
{
	__FuncGuard("CGLState::ShadeModel");

	if(m_p_cur_state->n_shade_model != n_model)
		glShadeModel(m_p_cur_state->n_shade_model = n_model);
}

void __fastcall CGLState::LightAmbient(int n_light, const GLfloat *p_color4)
{
	__FuncGuard("CGLState::LightAmbient");

	_ASSERTE(n_light >= 0 && n_light < m_p_cur_state->n_max_lights);

	if(memcmp(m_p_cur_state->t_light_data[n_light].f_ambient, p_color4, 4 * sizeof(GLfloat))) {
		memcpy(m_p_cur_state->t_light_data[n_light].f_ambient, p_color4, 4 * sizeof(GLfloat));
		glLightfv(GL_LIGHT0 + n_light, GL_AMBIENT, p_color4);
	}
}

void __fastcall CGLState::LightDiffuse(int n_light, const GLfloat *p_color4)
{
	__FuncGuard("CGLState::LightDiffuse");

	_ASSERTE(n_light >= 0 && n_light < m_p_cur_state->n_max_lights);

	if(memcmp(m_p_cur_state->t_light_data[n_light].f_diffuse, p_color4, 4 * sizeof(GLfloat))) {
		memcpy(m_p_cur_state->t_light_data[n_light].f_diffuse, p_color4, 4 * sizeof(GLfloat));
		glLightfv(GL_LIGHT0 + n_light, GL_DIFFUSE, p_color4);
	}
}

void __fastcall CGLState::LightSpecular(int n_light, const GLfloat *p_color4)
{
	__FuncGuard("CGLState::LightSpecular");

	_ASSERTE(n_light >= 0 && n_light < m_p_cur_state->n_max_lights);

	if(memcmp(m_p_cur_state->t_light_data[n_light].f_specular, p_color4, 4 * sizeof(GLfloat))) {
		memcpy(m_p_cur_state->t_light_data[n_light].f_specular, p_color4, 4 * sizeof(GLfloat));
		glLightfv(GL_LIGHT0 + n_light, GL_SPECULAR, p_color4);
	}
}

void __fastcall CGLState::LightPosition(int n_light, const GLfloat *p_position4)
{
	__FuncGuard("CGLState::LightPosition");

	_ASSERTE(n_light >= 0 && n_light < m_p_cur_state->n_max_lights);

	if(memcmp(m_p_cur_state->t_light_data[n_light].f_position, p_position4, 4 * sizeof(GLfloat))) {
		memcpy(m_p_cur_state->t_light_data[n_light].f_position, p_position4, 4 * sizeof(GLfloat));
		glLightfv(GL_LIGHT0 + n_light, GL_POSITION, p_position4);
	}
}

void __fastcall CGLState::LightSpotDirection(int n_light, const GLfloat *p_direction3)
{
	__FuncGuard("CGLState::LightSpotDirection");

	_ASSERTE(n_light >= 0 && n_light < m_p_cur_state->n_max_lights);

	if(memcmp(m_p_cur_state->t_light_data[n_light].f_spot_direction, p_direction3, 3 * sizeof(GLfloat))) {
		memcpy(m_p_cur_state->t_light_data[n_light].f_spot_direction, p_direction3, 3 * sizeof(GLfloat));
		glLightfv(GL_LIGHT0 + n_light, GL_SPOT_DIRECTION, p_direction3);
	}
}

void __fastcall CGLState::LightSpotExponent(int n_light, GLfloat f_value)
{
	__FuncGuard("CGLState::LightSpotExponent");

	_ASSERTE(n_light >= 0 && n_light < m_p_cur_state->n_max_lights);

	if(m_p_cur_state->t_light_data[n_light].f_spot_exponent != f_value) {
		glLightf(GL_LIGHT0 + n_light, GL_SPOT_EXPONENT,
			m_p_cur_state->t_light_data[n_light].f_spot_exponent = f_value);
	}
}

void __fastcall CGLState::LightSpotCutoff(int n_light, GLfloat f_value)
{
	__FuncGuard("CGLState::LightSpotCutoff");

	_ASSERTE(n_light >= 0 && n_light < m_p_cur_state->n_max_lights);

	if(m_p_cur_state->t_light_data[n_light].f_spot_cutoff != f_value) {
		glLightf(GL_LIGHT0 + n_light, GL_SPOT_CUTOFF,
			m_p_cur_state->t_light_data[n_light].f_spot_cutoff = f_value);
	}
}

void __fastcall CGLState::LightConstantAttenuation(int n_light, GLfloat f_value)
{
	__FuncGuard("CGLState::LightConstantAttenuation");

	_ASSERTE(n_light >= 0 && n_light < m_p_cur_state->n_max_lights);

	if(m_p_cur_state->t_light_data[n_light].f_constant_attenuation != f_value) {
		glLightf(GL_LIGHT0 + n_light, GL_CONSTANT_ATTENUATION,
			m_p_cur_state->t_light_data[n_light].f_constant_attenuation = f_value);
	}
}

void __fastcall CGLState::LightLinearAttenuation(int n_light, GLfloat f_value)
{
	__FuncGuard("CGLState::LightLinearAttenuation");

	_ASSERTE(n_light >= 0 && n_light < m_p_cur_state->n_max_lights);

	if(m_p_cur_state->t_light_data[n_light].f_linear_attenuation != f_value) {
		glLightf(GL_LIGHT0 + n_light, GL_LINEAR_ATTENUATION,
			m_p_cur_state->t_light_data[n_light].f_linear_attenuation = f_value);
	}
}

void __fastcall CGLState::LightQuadraticAttenuation(int n_light, GLfloat f_value)
{
	__FuncGuard("CGLState::LightQuadraticAttenuation");

	_ASSERTE(n_light >= 0 && n_light < m_p_cur_state->n_max_lights);

	if(m_p_cur_state->t_light_data[n_light].f_quadratic_attenuation != f_value) {
		glLightf(GL_LIGHT0 + n_light, GL_QUADRATIC_ATTENUATION,
			m_p_cur_state->t_light_data[n_light].f_quadratic_attenuation = f_value);
	}
}

void __fastcall CGLState::EnableBlend()
{
	__FuncGuard("CGLState::EnableBlend");

	if(!m_p_cur_state->b_blend) {
		glEnable(GL_BLEND);
		m_p_cur_state->b_blend = true;
	}
}

void __fastcall CGLState::DisableBlend()
{
	__FuncGuard("CGLState::DisableBlend");

	if(m_p_cur_state->b_blend) {
		glDisable(GL_BLEND);
		m_p_cur_state->b_blend = false;
	}
}

void __fastcall CGLState::BlendFunc(GLenum n_src_op, GLenum n_dest_op)
{
	__FuncGuard("CGLState::BlendFunc");

	if(m_p_cur_state->n_blend_func[0] != n_src_op ||
	   m_p_cur_state->n_blend_func[1] != n_dest_op) {
		glBlendFunc(m_p_cur_state->n_blend_func[0] = n_src_op,
			m_p_cur_state->n_blend_func[1] = n_dest_op);
	}
}

void __fastcall CGLState::EnableDepthTest()
{
	__FuncGuard("CGLState::EnableDepthTest");

	if(!m_p_cur_state->b_depth_test) {
		glEnable(GL_DEPTH_TEST);
		m_p_cur_state->b_depth_test = true;
	}
}

void __fastcall CGLState::DisableDepthTest()
{
	__FuncGuard("CGLState::DisableDepthTest");

	if(m_p_cur_state->b_depth_test) {
		glDisable(GL_DEPTH_TEST);
		m_p_cur_state->b_depth_test = false;
	}
}

void __fastcall CGLState::DepthFunc(GLenum n_op)
{
	__FuncGuard("CGLState::DepthFunc");

	if(m_p_cur_state->n_depth_func != n_op)
		glDepthFunc(m_p_cur_state->n_depth_func = n_op);
}

void __fastcall CGLState::EnableStencilTest()
{
	__FuncGuard("CGLState::EnableStencilTest");

	if(!m_p_cur_state->b_stencil_test) {
		m_p_cur_state->b_stencil_test = true;
		glEnable(GL_STENCIL_TEST);
	}
}

void __fastcall CGLState::DisableStencilTest()
{
	__FuncGuard("CGLState::DisableStencilTest");

	if(m_p_cur_state->b_stencil_test) {
		m_p_cur_state->b_stencil_test = false;
		glDisable(GL_STENCIL_TEST);
	}
}

void __fastcall CGLState::StencilFunc(GLenum n_func, GLint n_ref, GLuint n_mask)
{
	__FuncGuard("CGLState::StencilFunc");

	if(m_p_cur_state->n_stencil_func[0] != n_func ||
	   m_p_cur_state->n_stencil_func[1] != n_func ||
	   m_p_cur_state->n_stencil_ref[0] != n_ref ||
	   m_p_cur_state->n_stencil_ref[1] != n_ref ||
	   m_p_cur_state->n_stencil_mask[0] != n_mask ||
	   m_p_cur_state->n_stencil_mask[1] != n_mask) {
		if(m_p_cur_state->b_separate_gl2_stencil) {
			glStencilFuncSeparate(GL_FRONT_AND_BACK,
				m_p_cur_state->n_stencil_func[0] = m_p_cur_state->n_stencil_func[1] = n_func,
				m_p_cur_state->n_stencil_ref[0] = m_p_cur_state->n_stencil_ref[1] = n_ref,
				m_p_cur_state->n_stencil_mask[0] = m_p_cur_state->n_stencil_mask[1] = n_mask);
		} else {
			glStencilFunc(m_p_cur_state->n_stencil_func[0] =
				m_p_cur_state->n_stencil_func[1] = n_func,
				m_p_cur_state->n_stencil_ref[0] =
				m_p_cur_state->n_stencil_ref[1] = n_ref,
				m_p_cur_state->n_stencil_mask[0] =
				m_p_cur_state->n_stencil_mask[1] = n_mask);
		}
	}
}

void __fastcall CGLState::StencilFunc(bool b_front, GLenum n_func, GLint n_ref, GLuint n_mask)
{
	__FuncGuard("CGLState::StencilFunc");

	_ASSERTE(b_front == !!b_front);
	if(m_p_cur_state->n_stencil_func[b_front] != n_func ||
	   m_p_cur_state->n_stencil_ref[b_front] != n_ref ||
	   m_p_cur_state->n_stencil_mask[b_front] != n_mask) {
		_ASSERTE(m_p_cur_state->b_separate_gl2_stencil);
		glStencilFuncSeparate((b_front)? GL_FRONT : GL_BACK,
			m_p_cur_state->n_stencil_func[b_front] = n_func,
			m_p_cur_state->n_stencil_ref[b_front] = n_ref,
			m_p_cur_state->n_stencil_mask[b_front] = n_mask);
	}
}

void __fastcall CGLState::StencilOp(GLenum n_s_fail, GLint n_s_pass_d_fail, GLuint n_s_pass_d_pass)
{
	__FuncGuard("CGLState::StencilOp");

	if(m_p_cur_state->n_stencil_op_sfail[0] != n_s_fail ||
	   m_p_cur_state->n_stencil_op_sfail[1] != n_s_fail ||
	   m_p_cur_state->n_stencil_op_dfail[0] != n_s_pass_d_fail ||
	   m_p_cur_state->n_stencil_op_dfail[1] != n_s_pass_d_fail ||
	   m_p_cur_state->n_stencil_op_dpass[0] != n_s_pass_d_pass ||
	   m_p_cur_state->n_stencil_op_dpass[1] != n_s_pass_d_pass) {
		if(m_p_cur_state->b_separate_gl2_stencil) {
			glStencilOpSeparate(GL_FRONT_AND_BACK,
				m_p_cur_state->n_stencil_op_sfail[0] =
				m_p_cur_state->n_stencil_op_sfail[1] = n_s_fail,
				m_p_cur_state->n_stencil_op_dfail[0] =
				m_p_cur_state->n_stencil_op_dfail[1] = n_s_pass_d_fail,
				m_p_cur_state->n_stencil_op_dpass[0] =
				m_p_cur_state->n_stencil_op_dpass[1] = n_s_pass_d_pass);
		} else {
			glStencilOp(m_p_cur_state->n_stencil_op_sfail[0] =
				m_p_cur_state->n_stencil_op_sfail[1] = n_s_fail,
				m_p_cur_state->n_stencil_op_dfail[0] =
				m_p_cur_state->n_stencil_op_dfail[1] = n_s_pass_d_fail,
				m_p_cur_state->n_stencil_op_dpass[0] =
				m_p_cur_state->n_stencil_op_dpass[1] = n_s_pass_d_pass);
		}
	}
}

void __fastcall CGLState::StencilOp(bool b_front, GLenum n_s_fail, GLint n_s_pass_d_fail, GLuint n_s_pass_d_pass)
{
	__FuncGuard("CGLState::StencilOp");

	if(m_p_cur_state->n_stencil_op_sfail[b_front] != n_s_fail ||
	   m_p_cur_state->n_stencil_op_dfail[b_front] != n_s_pass_d_fail ||
	   m_p_cur_state->n_stencil_op_dpass[b_front] != n_s_pass_d_pass) {
		_ASSERTE(m_p_cur_state->b_separate_gl2_stencil);
		glStencilOpSeparate((b_front)? GL_FRONT : GL_BACK,
			m_p_cur_state->n_stencil_op_sfail[b_front] = n_s_fail,
			m_p_cur_state->n_stencil_op_dfail[b_front] = n_s_pass_d_fail,
			m_p_cur_state->n_stencil_op_dpass[b_front] = n_s_pass_d_pass);
	}
}

void __fastcall CGLState::EnableAlphaTest()
{
	__FuncGuard("CGLState::EnableAlphaTest");

	if(!m_p_cur_state->b_alpha_test) {
		glEnable(GL_ALPHA_TEST);
		m_p_cur_state->b_alpha_test = true;
	}
}

void __fastcall CGLState::DisableAlphaTest()
{
	__FuncGuard("CGLState::DisableAlphaTest");

	if(m_p_cur_state->b_alpha_test) {
		glDisable(GL_ALPHA_TEST);
		m_p_cur_state->b_alpha_test = false;
	}
}

void __fastcall CGLState::AlphaFunc(GLenum n_op, GLclampf f_tresh)
{
	__FuncGuard("CGLState::AlphaFunc");

	if(m_p_cur_state->n_depth_func != n_op ||
	   m_p_cur_state->f_alpha_tresh != f_tresh) {
		glAlphaFunc(m_p_cur_state->n_depth_func = n_op,
			m_p_cur_state->f_alpha_tresh = f_tresh);
	}
}

void __fastcall CGLState::DisableAllTexture1D()
{
	__FuncGuard("CGLState::DisableAllTexture1D");

	int i;

	for(i = 0; i < m_p_cur_state->n_max_texture_units; i ++) {
#ifdef __MULTITEXTURE_ENABLED
		glActiveTexture(GL_TEXTURE0 + i);
#endif
		if(m_p_cur_state->b_texture1d_enabled[i]) {
			glDisable(GL_TEXTURE_1D);
			m_p_cur_state->b_texture1d_enabled[i] = false;
		}
	}
#ifdef __MULTITEXTURE_ENABLED
	glActiveTexture(GL_TEXTURE0 + m_p_cur_state->n_active_tex_unit);
#endif
}

void __fastcall CGLState::DisableAllTextures()
{
	int i;

	for(i = 0; i < m_p_cur_state->n_max_texture_units; i ++) {
#ifdef __MULTITEXTURE_ENABLED
		glActiveTexture(GL_TEXTURE0 + i);
#endif
		if(m_p_cur_state->b_texture1d_enabled[i]) {
			glDisable(GL_TEXTURE_1D);
			m_p_cur_state->b_texture1d_enabled[i] = false;
		}
		if(m_p_cur_state->b_texture2d_enabled[i]) {
			glDisable(GL_TEXTURE_2D);
			m_p_cur_state->b_texture2d_enabled[i] = false;
		}
		if(m_p_cur_state->b_texture3d_enabled[i]) {
			glDisable(GL_TEXTURE_3D);
			m_p_cur_state->b_texture3d_enabled[i] = false;
		}
		if(m_p_cur_state->b_texture_cubemap_enabled[i]) {
			glDisable(GL_TEXTURE_CUBE_MAP);
			m_p_cur_state->b_texture_cubemap_enabled[i] = false;
		}
	}
#ifdef __MULTITEXTURE_ENABLED
	glActiveTexture(GL_TEXTURE0 + m_p_cur_state->n_active_tex_unit);
#endif
}

void __fastcall CGLState::DisableAllTexture2D()
{
	__FuncGuard("CGLState::DisableAllTexture2D");

	int i;

	for(i = 0; i < m_p_cur_state->n_max_texture_units; i ++) {
#ifdef __MULTITEXTURE_ENABLED
		glActiveTexture(GL_TEXTURE0 + i);
#endif
		if(m_p_cur_state->b_texture2d_enabled[i]) {
			glDisable(GL_TEXTURE_2D);
			m_p_cur_state->b_texture2d_enabled[i] = false;
		}
	}
#ifdef __MULTITEXTURE_ENABLED
	glActiveTexture(GL_TEXTURE0 + m_p_cur_state->n_active_tex_unit);
#endif
}

void __fastcall CGLState::DisableAllTexture3D()
{
	__FuncGuard("CGLState::DisableAllTexture3D");

	int i;

	for(i = 0; i < m_p_cur_state->n_max_texture_units; i ++) {
#ifdef __MULTITEXTURE_ENABLED
		glActiveTexture(GL_TEXTURE0 + i);
#endif
		if(m_p_cur_state->b_texture3d_enabled[i]) {
			glDisable(GL_TEXTURE_3D);
			m_p_cur_state->b_texture3d_enabled[i] = false;
		}
	}
#ifdef __MULTITEXTURE_ENABLED
	glActiveTexture(GL_TEXTURE0 + m_p_cur_state->n_active_tex_unit);
#endif
}

void __fastcall CGLState::DisableAllTextureCubeMap()
{
	__FuncGuard("CGLState::DisableAllTextureCubeMap");

	int i;

	for(i = 0; i < m_p_cur_state->n_max_texture_units; i ++) {
#ifdef __MULTITEXTURE_ENABLED
		glActiveTexture(GL_TEXTURE0 + i);
#endif
		if(m_p_cur_state->b_texture_cubemap_enabled[i]) {
			glDisable(GL_TEXTURE_CUBE_MAP);
			m_p_cur_state->b_texture_cubemap_enabled[i] = false;
		}
	}
#ifdef __MULTITEXTURE_ENABLED
	glActiveTexture(GL_TEXTURE0 + m_p_cur_state->n_active_tex_unit);
#endif
}

void __fastcall CGLState::DisableAllTexture(GLenum n_type)
{
	__FuncGuard("CGLState::DisableAllTexture");

	switch(n_type) {
	case GL_TEXTURE_1D:
		DisableAllTexture1D();
		return;
	case GL_TEXTURE_2D:
		DisableAllTexture2D();
		return;
	case GL_TEXTURE_3D:
		DisableAllTexture3D();
		return;
	case GL_TEXTURE_CUBE_MAP:
		DisableAllTextureCubeMap();
		return;
	default:
		for(int i = 0; i < m_p_cur_state->n_max_texture_units; i ++) {
#ifdef __MULTITEXTURE_ENABLED
			glActiveTexture(GL_TEXTURE0 + i);
#endif
			glDisable(n_type);
		}
#ifdef __MULTITEXTURE_ENABLED
		glActiveTexture(m_p_cur_state->n_active_tex_unit);
#endif
	}
}

void __fastcall CGLState::EnableTexture1D()
{
	__FuncGuard("CGLState::EnableTexture1D");

	if(!m_p_cur_state->b_texture1d_enabled[m_p_cur_state->n_active_tex_unit]) {
		glEnable(GL_TEXTURE_1D);
		m_p_cur_state->b_texture1d_enabled[m_p_cur_state->n_active_tex_unit] = true;
	}
}

void __fastcall CGLState::EnableTexture2D()
{
	__FuncGuard("CGLState::EnableTexture2D");

	if(!m_p_cur_state->b_texture2d_enabled[m_p_cur_state->n_active_tex_unit]) {
		glEnable(GL_TEXTURE_2D);
		m_p_cur_state->b_texture2d_enabled[m_p_cur_state->n_active_tex_unit] = true;
	}
}

void __fastcall CGLState::EnableTexture3D()
{
	__FuncGuard("CGLState::EnableTexture3D");

	if(!m_p_cur_state->b_texture3d_enabled[m_p_cur_state->n_active_tex_unit]) {
		glEnable(GL_TEXTURE_3D);
		m_p_cur_state->b_texture3d_enabled[m_p_cur_state->n_active_tex_unit] = true;
	}
}

void __fastcall CGLState::EnableTextureCubeMap()
{
	__FuncGuard("CGLState::EnableTextureCubeMap");

	if(!m_p_cur_state->b_texture_cubemap_enabled[m_p_cur_state->n_active_tex_unit]) {
		glEnable(GL_TEXTURE_CUBE_MAP);
		m_p_cur_state->b_texture_cubemap_enabled[m_p_cur_state->n_active_tex_unit] = true;
	}
}

void __fastcall CGLState::EnableTexture(GLenum n_type)
{
	__FuncGuard("CGLState::EnableTexture");

	switch(n_type) {
	case GL_TEXTURE_1D:
		EnableTexture1D();
		return;
	case GL_TEXTURE_2D:
		EnableTexture2D();
		return;
	case GL_TEXTURE_3D:
		EnableTexture3D();
		return;
	case GL_TEXTURE_CUBE_MAP:
		EnableTextureCubeMap();
		return;
	default:
		glEnable(n_type);
	}
}

void __fastcall CGLState::DisableTexture1D()
{
	__FuncGuard("CGLState::DisableTexture1D");

	if(m_p_cur_state->b_texture1d_enabled[m_p_cur_state->n_active_tex_unit]) {
		glDisable(GL_TEXTURE_1D);
		m_p_cur_state->b_texture1d_enabled[m_p_cur_state->n_active_tex_unit] = false;
	}
}

void __fastcall CGLState::DisableTexture2D()
{
	__FuncGuard("CGLState::DisableTexture2D");

	if(m_p_cur_state->b_texture2d_enabled[m_p_cur_state->n_active_tex_unit]) {
		glDisable(GL_TEXTURE_2D);
		m_p_cur_state->b_texture2d_enabled[m_p_cur_state->n_active_tex_unit] = false;
	}
}

void __fastcall CGLState::DisableTexture3D()
{
	__FuncGuard("CGLState::DisableTexture3D");

	if(m_p_cur_state->b_texture3d_enabled[m_p_cur_state->n_active_tex_unit]) {
		glDisable(GL_TEXTURE_3D);
		m_p_cur_state->b_texture3d_enabled[m_p_cur_state->n_active_tex_unit] = false;
	}
}

void __fastcall CGLState::DisableTextureCubeMap()
{
	__FuncGuard("CGLState::DisableTextureCubeMap");

	if(m_p_cur_state->b_texture_cubemap_enabled[m_p_cur_state->n_active_tex_unit]) {
		glDisable(GL_TEXTURE_CUBE_MAP);
		m_p_cur_state->b_texture_cubemap_enabled[m_p_cur_state->n_active_tex_unit] = false;
	}
}

void __fastcall CGLState::DisableTexture(GLenum n_type)
{
	__FuncGuard("CGLState::DisableTexture");

	switch(n_type) {
	case GL_TEXTURE_1D:
		DisableTexture1D();
		return;
	case GL_TEXTURE_2D:
		DisableTexture2D();
		return;
	case GL_TEXTURE_3D:
		DisableTexture3D();
		return;
	case GL_TEXTURE_CUBE_MAP:
		DisableTextureCubeMap();
		return;
	default:
		glDisable(n_type);
	}
}

void __fastcall CGLState::ActiveTextureUnit(int n_unit)
{
	__FuncGuard("CGLState::ActiveTextureUnit");

	_ASSERTE(n_unit >= 0 && n_unit < m_p_cur_state->n_max_texture_units);

#ifdef __MULTITEXTURE_ENABLED
	if(m_p_cur_state->n_active_tex_unit != n_unit)
		glActiveTexture(GL_TEXTURE0 + (m_p_cur_state->n_active_tex_unit = n_unit));
#endif
}

int __fastcall CGLState::n_TexUnit_Num()
{
	__FuncGuard("CGLState::n_TexUnit_Num");

	return m_p_cur_state->n_max_texture_units;
}

void __fastcall CGLState::BindTexture1D(GLuint n_texture)
{
	__FuncGuard("CGLState::BindTexture1D");

	if(m_p_cur_state->n_texture1d[m_p_cur_state->n_active_tex_unit] != n_texture) {
		glBindTexture(GL_TEXTURE_1D,
			m_p_cur_state->n_texture1d[m_p_cur_state->n_active_tex_unit] = n_texture);
	}
}

void __fastcall CGLState::BindTexture2D(GLuint n_texture)
{
	__FuncGuard("CGLState::BindTexture2D");

	if(m_p_cur_state->n_texture2d[m_p_cur_state->n_active_tex_unit] != n_texture) {
		glBindTexture(GL_TEXTURE_2D,
			m_p_cur_state->n_texture2d[m_p_cur_state->n_active_tex_unit] = n_texture);
	}
}

void __fastcall CGLState::BindTexture3D(GLuint n_texture)
{
	__FuncGuard("CGLState::BindTexture3D");

	if(m_p_cur_state->n_texture3d[m_p_cur_state->n_active_tex_unit] != n_texture) {
		glBindTexture(GL_TEXTURE_3D,
			m_p_cur_state->n_texture3d[m_p_cur_state->n_active_tex_unit] = n_texture);
	}
}

void __fastcall CGLState::BindTextureCubeMap(GLuint n_texture)
{
	__FuncGuard("CGLState::BindTextureCubeMap");

	if(m_p_cur_state->n_texture_cubemap[m_p_cur_state->n_active_tex_unit] != n_texture) {
		glBindTexture(GL_TEXTURE_CUBE_MAP,
			m_p_cur_state->n_texture_cubemap[m_p_cur_state->n_active_tex_unit] = n_texture);
	}
}

void __fastcall CGLState::BindTexture(GLenum n_type, GLuint n_texture)
{
	__FuncGuard("CGLState::BindTexture");

	switch(n_type) {
	case GL_TEXTURE_1D:
		BindTexture1D(n_texture);
		return;
	case GL_TEXTURE_2D:
		BindTexture2D(n_texture);
		return;
	case GL_TEXTURE_3D:
		BindTexture3D(n_texture);
		return;
	case GL_TEXTURE_CUBE_MAP:
		BindTextureCubeMap(n_texture);
		return;
	default:
		glBindTexture(n_type, n_texture);
	}
}

void __fastcall CGLState::Color3f(GLfloat r, GLfloat g, GLfloat b)
{
	__FuncGuard("CGLState::Color3f");

	if(m_p_cur_state->f_color[0] != r ||
	   m_p_cur_state->f_color[1] != g ||
	   m_p_cur_state->f_color[2] != b/* ||
	   m_p_cur_state->f_color[3] != 1.0f*/) {
		/*if(m_p_cur_state->f_color[3] != 1.0f) {
			glColor4f(m_p_cur_state->f_color[0] = r,
					  m_p_cur_state->f_color[1] = g,
					  m_p_cur_state->f_color[2] = b, 1.0f);
			m_p_cur_state->f_color[3] = 1.0f;
			m_p_cur_state->n_color[3] = 0xff;
		} else {*/
			glColor3f(m_p_cur_state->f_color[0] = r,
					  m_p_cur_state->f_color[1] = g,
					  m_p_cur_state->f_color[2] = b);
		/*}*/
		//
		m_p_cur_state->n_color[0] = (GLubyte)(m_p_cur_state->f_color[0] * 0xff);
		m_p_cur_state->n_color[1] = (GLubyte)(m_p_cur_state->f_color[1] * 0xff);
		m_p_cur_state->n_color[2] = (GLubyte)(m_p_cur_state->f_color[2] * 0xff);
	}
}

void __fastcall CGLState::Color4f(GLfloat r, GLfloat g, GLfloat b, GLfloat a)
{
	__FuncGuard("CGLState::Color4f");

	if(m_p_cur_state->f_color[0] != r ||
	   m_p_cur_state->f_color[1] != g ||
	   m_p_cur_state->f_color[2] != b ||
	   m_p_cur_state->f_color[3] != a) {
		glColor4f(m_p_cur_state->f_color[0] = r,
				  m_p_cur_state->f_color[1] = g,
				  m_p_cur_state->f_color[2] = b,
				  m_p_cur_state->f_color[3] = a);
		//
		m_p_cur_state->n_color[0] = (GLubyte)(m_p_cur_state->f_color[0] * 0xff);
		m_p_cur_state->n_color[1] = (GLubyte)(m_p_cur_state->f_color[1] * 0xff);
		m_p_cur_state->n_color[2] = (GLubyte)(m_p_cur_state->f_color[2] * 0xff);
		m_p_cur_state->n_color[3] = (GLubyte)(m_p_cur_state->f_color[3] * 0xff);
	}
}

void __fastcall CGLState::Color3b(GLubyte r, GLubyte g, GLubyte b)
{
	__FuncGuard("CGLState::Color3b");

	if(m_p_cur_state->n_color[0] != r ||
	   m_p_cur_state->n_color[1] != g ||
	   m_p_cur_state->n_color[2] != b/* ||
	   m_p_cur_state->n_color[2] != 0xff*/) {
		/*if(m_p_cur_state->n_color[2] != 0xff) {
			glColor4b(m_p_cur_state->n_color[0] = r,
					  m_p_cur_state->n_color[1] = g,
					  m_p_cur_state->n_color[2] = b, 0xff);
			m_p_cur_state->f_color[3] = 1.0f;
			m_p_cur_state->n_color[3] = 0xff;
		} else {*/
			glColor3b(m_p_cur_state->n_color[0] = r,
					  m_p_cur_state->n_color[1] = g,
					  m_p_cur_state->n_color[2] = b);
		/*}*/
		//
		m_p_cur_state->f_color[0] = m_p_cur_state->f_color[0] / 255.0f;
		m_p_cur_state->f_color[1] = m_p_cur_state->f_color[1] / 255.0f;
		m_p_cur_state->f_color[2] = m_p_cur_state->f_color[2] / 255.0f;
	}
}

void __fastcall CGLState::Color4b(GLubyte r, GLubyte g, GLubyte b, GLubyte a)
{
	__FuncGuard("CGLState::Color4b");

	if(m_p_cur_state->n_color[0] != r ||
	   m_p_cur_state->n_color[1] != g ||
	   m_p_cur_state->n_color[2] != b ||
	   m_p_cur_state->n_color[3] != a) {
		glColor4b(m_p_cur_state->n_color[0] = r,
				  m_p_cur_state->n_color[1] = g,
				  m_p_cur_state->n_color[2] = b,
				  m_p_cur_state->n_color[3] = a);
		//
		m_p_cur_state->f_color[0] = m_p_cur_state->f_color[0] / 255.0f;
		m_p_cur_state->f_color[1] = m_p_cur_state->f_color[1] / 255.0f;
		m_p_cur_state->f_color[2] = m_p_cur_state->f_color[2] / 255.0f;
		m_p_cur_state->f_color[3] = m_p_cur_state->f_color[3] / 255.0f;
	}
}

void __fastcall CGLState::ClearColor4f(GLclampf r, GLclampf g, GLclampf b, GLclampf a)
{
	__FuncGuard("CGLState::ClearColor4f");

	if(m_p_cur_state->f_clr_color[0] != r ||
	   m_p_cur_state->f_clr_color[1] != g ||
	   m_p_cur_state->f_clr_color[2] != b ||
	   m_p_cur_state->f_clr_color[3] != a) {
		glClearColor(m_p_cur_state->f_clr_color[0] = r,
					 m_p_cur_state->f_clr_color[1] = g,
					 m_p_cur_state->f_clr_color[2] = b,
					 m_p_cur_state->f_clr_color[3] = a);
	}
}

void __fastcall CGLState::ClearDepth(GLclampd d)
{
	__FuncGuard("CGLState::ClearDepth");

	if(m_p_cur_state->f_clr_depth != d)
		glClearDepth(m_p_cur_state->f_clr_depth = d);
}

void __fastcall CGLState::ClearStencil(GLint s)
{
	__FuncGuard("CGLState::ClearStencil");

	if(m_p_cur_state->n_clr_stencil != s)
		glClearStencil(m_p_cur_state->n_clr_stencil = s);
}

void __fastcall CGLState::EnableCullFace()
{
	__FuncGuard("CGLState::EnableCullFace");

	if(!m_p_cur_state->b_cull_face) {
		glEnable(GL_CULL_FACE);
		m_p_cur_state->b_cull_face = true;
	}
}

void __fastcall CGLState::DisableCullFace()
{
	__FuncGuard("CGLState::DisableCullFace");

	if(m_p_cur_state->b_cull_face) {
		glDisable(GL_CULL_FACE);
		m_p_cur_state->b_cull_face = false;
	}
}

void __fastcall CGLState::CullFace(GLenum n_mode)
{
	__FuncGuard("CGLState::CullFace");

	if(m_p_cur_state->n_cull_face != n_mode)
		glCullFace(m_p_cur_state->n_cull_face = n_mode);
}

void __fastcall CGLState::FrontFace(GLenum n_mode)
{
	__FuncGuard("CGLState::FrontFace");

	if(m_p_cur_state->n_front_face != n_mode)
		glFrontFace(m_p_cur_state->n_front_face = n_mode);
}

void __fastcall CGLState::Enable_STRQ_TexGen()
{
	__FuncGuard("CGLState::Enable_STRQ_TexGen");

	if(!m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][0])
		glEnable(GL_TEXTURE_GEN_S);
	if(!m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][1])
		glEnable(GL_TEXTURE_GEN_T);
	if(!m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][2])
		glEnable(GL_TEXTURE_GEN_R);
	if(!m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][3])
		glEnable(GL_TEXTURE_GEN_Q);
	//
	m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][0] =
		m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][1] =
		m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][2] =
		m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][3] = true;
}

void __fastcall CGLState::Disable_STRQ_Texgen()
{
	__FuncGuard("CGLState::Disable_STRQ_Texgen");

	if(m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][0])
		glDisable(GL_TEXTURE_GEN_S);
	if(m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][1])
		glDisable(GL_TEXTURE_GEN_T);
	if(m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][2])
		glDisable(GL_TEXTURE_GEN_R);
	if(m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][3])
		glDisable(GL_TEXTURE_GEN_Q);
	//
	m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][0] =
		m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][1] =
		m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][2] =
		m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][3] = false;
}

void __fastcall CGLState::Enable_S_TexGen()
{
	__FuncGuard("CGLState::Enable_S_TexGen");

	if(!m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][0]) {
		glEnable(GL_TEXTURE_GEN_S);
		m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][0] = true;
	}
}

void __fastcall CGLState::Enable_T_TexGen()
{
	__FuncGuard("CGLState::Enable_T_TexGen");

	if(!m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][1]) {
		glEnable(GL_TEXTURE_GEN_T);
		m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][1] = true;
	}
}

void __fastcall CGLState::Enable_R_TexGen()
{
	__FuncGuard("CGLState::Enable_R_TexGen");

	if(!m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][2]) {
		glEnable(GL_TEXTURE_GEN_R);
		m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][2] = true;
	}
}

void __fastcall CGLState::Enable_Q_TexGen()
{
	__FuncGuard("CGLState::Enable_Q_TexGen");

	if(!m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][3]) {
		glEnable(GL_TEXTURE_GEN_Q);
		m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][3] = true;
	}
}

void __fastcall CGLState::Disable_S_TexGen()
{
	__FuncGuard("CGLState::Disable_S_TexGen");

	if(m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][0]) {
		glDisable(GL_TEXTURE_GEN_S);
		m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][0] = false;
	}
}

void __fastcall CGLState::Disable_T_TexGen()
{
	__FuncGuard("CGLState::Disable_T_TexGen");

	if(m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][1]) {
		glDisable(GL_TEXTURE_GEN_T);
		m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][1] = false;
	}
}

void __fastcall CGLState::Disable_R_TexGen()
{
	__FuncGuard("CGLState::Disable_R_TexGen");

	if(m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][2]) {
		glDisable(GL_TEXTURE_GEN_R);
		m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][2] = false;
	}
}

void __fastcall CGLState::Disable_Q_TexGen()
{
	__FuncGuard("CGLState::Disable_Q_TexGen");

	if(m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][3]) {
		glDisable(GL_TEXTURE_GEN_Q);
		m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][3] = false;
	}
}

void __fastcall CGLState::Enable_ST_TexGen()
{
	__FuncGuard("CGLState::Enable_ST_TexGen");

	if(!m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][0]) {
		glEnable(GL_TEXTURE_GEN_S);
		m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][0] = true;
	}
	if(!m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][1]) {
		glEnable(GL_TEXTURE_GEN_T);
		m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][1] = true;
	}
}

void __fastcall CGLState::Enable_STR_TexGen()
{
	__FuncGuard("CGLState::Enable_STR_TexGen");

	if(!m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][0]) {
		glEnable(GL_TEXTURE_GEN_S);
		m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][0] = true;
	}
	if(!m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][1]) {
		glEnable(GL_TEXTURE_GEN_T);
		m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][1] = true;
	}
	if(!m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][2]) {
		glEnable(GL_TEXTURE_GEN_R);
		m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][2] = true;
	}
}

void __fastcall CGLState::EnableTexGen(int n_coord)
{
	__FuncGuard("CGLState::EnableTexGen");

	if(!m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][n_coord]) {
		glEnable(n_coord + GL_TEXTURE_GEN_S);
		m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][n_coord] = true;
	}
}

void __fastcall CGLState::Disable_ST_TexGen()
{
	__FuncGuard("CGLState::Disable_ST_TexGen");

	if(m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][0]) {
		glDisable(GL_TEXTURE_GEN_S);
		m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][0] = false;
	}
	if(m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][1]) {
		glDisable(GL_TEXTURE_GEN_T);
		m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][1] = false;
	}
}

void __fastcall CGLState::Disable_STR_TexGen()
{
	__FuncGuard("CGLState::Disable_STR_TexGen");

	if(m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][0]) {
		glDisable(GL_TEXTURE_GEN_S);
		m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][0] = false;
	}
	if(m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][1]) {
		glDisable(GL_TEXTURE_GEN_T);
		m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][1] = false;
	}
	if(m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][2]) {
		glDisable(GL_TEXTURE_GEN_R);
		m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][2] = false;
	}
}

void __fastcall CGLState::DisableTexGen(int n_coord)
{
	__FuncGuard("CGLState::DisableTexGen");

	if(m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][n_coord]) {
		glDisable(n_coord + GL_TEXTURE_GEN_S);
		m_p_cur_state->b_texgen_enabled[m_p_cur_state->n_active_tex_unit][n_coord] = false;
	}
}

void __fastcall CGLState::TexGen_STRQ_fv(int n_name, GLfloat *p_param)
{
	__FuncGuard("CGLState::TexGen_STRQ_fv");

	if(n_name == GL_OBJECT_PLANE) {
		if(memcmp(m_p_cur_state->f_texgen_object_plane[m_p_cur_state->n_active_tex_unit][0], p_param, 4 * sizeof(GLfloat))) {
			memcpy(m_p_cur_state->f_texgen_object_plane[m_p_cur_state->n_active_tex_unit][0], p_param, 4 * sizeof(GLfloat));
			glTexGenfv(GL_S, GL_OBJECT_PLANE, p_param);
		}
		if(memcmp(m_p_cur_state->f_texgen_object_plane[m_p_cur_state->n_active_tex_unit][1], p_param, 4 * sizeof(GLfloat))) {
			memcpy(m_p_cur_state->f_texgen_object_plane[m_p_cur_state->n_active_tex_unit][1], p_param, 4 * sizeof(GLfloat));
			glTexGenfv(GL_T, GL_OBJECT_PLANE, p_param);
		}
		if(memcmp(m_p_cur_state->f_texgen_object_plane[m_p_cur_state->n_active_tex_unit][2], p_param, 4 * sizeof(GLfloat))) {
			memcpy(m_p_cur_state->f_texgen_object_plane[m_p_cur_state->n_active_tex_unit][2], p_param, 4 * sizeof(GLfloat));
			glTexGenfv(GL_R, GL_OBJECT_PLANE, p_param);
		}
		if(memcmp(m_p_cur_state->f_texgen_object_plane[m_p_cur_state->n_active_tex_unit][3], p_param, 4 * sizeof(GLfloat))) {
			memcpy(m_p_cur_state->f_texgen_object_plane[m_p_cur_state->n_active_tex_unit][3], p_param, 4 * sizeof(GLfloat));
			glTexGenfv(GL_Q, GL_OBJECT_PLANE, p_param);
		}
	} else if(n_name == GL_EYE_PLANE) {
		if(memcmp(m_p_cur_state->f_texgen_eye_plane[m_p_cur_state->n_active_tex_unit][0], p_param, 4 * sizeof(GLfloat))) {
			memcpy(m_p_cur_state->f_texgen_eye_plane[m_p_cur_state->n_active_tex_unit][0], p_param, 4 * sizeof(GLfloat));
			glTexGenfv(GL_S, GL_EYE_PLANE, p_param);
		}
		if(memcmp(m_p_cur_state->f_texgen_eye_plane[m_p_cur_state->n_active_tex_unit][1], p_param, 4 * sizeof(GLfloat))) {
			memcpy(m_p_cur_state->f_texgen_eye_plane[m_p_cur_state->n_active_tex_unit][1], p_param, 4 * sizeof(GLfloat));
			glTexGenfv(GL_T, GL_EYE_PLANE, p_param);
		}
		if(memcmp(m_p_cur_state->f_texgen_eye_plane[m_p_cur_state->n_active_tex_unit][2], p_param, 4 * sizeof(GLfloat))) {
			memcpy(m_p_cur_state->f_texgen_eye_plane[m_p_cur_state->n_active_tex_unit][2], p_param, 4 * sizeof(GLfloat));
			glTexGenfv(GL_R, GL_EYE_PLANE, p_param);
		}
		if(memcmp(m_p_cur_state->f_texgen_eye_plane[m_p_cur_state->n_active_tex_unit][3], p_param, 4 * sizeof(GLfloat))) {
			memcpy(m_p_cur_state->f_texgen_eye_plane[m_p_cur_state->n_active_tex_unit][3], p_param, 4 * sizeof(GLfloat));
			glTexGenfv(GL_Q, GL_EYE_PLANE, p_param);
		}
	} else {
		_ASSERTE(0);
	}
}

void __fastcall CGLState::TexGen_S_fv(int n_name, GLfloat *p_param)
{
	__FuncGuard("CGLState::TexGen_S_fv");

	if(n_name == GL_OBJECT_PLANE) {
		if(memcmp(m_p_cur_state->f_texgen_object_plane[m_p_cur_state->n_active_tex_unit][0], p_param, 4 * sizeof(GLfloat))) {
			memcpy(m_p_cur_state->f_texgen_object_plane[m_p_cur_state->n_active_tex_unit][0], p_param, 4 * sizeof(GLfloat));
			glTexGenfv(GL_S, GL_OBJECT_PLANE, p_param);
		}
	} else if(n_name == GL_EYE_PLANE) {
		if(memcmp(m_p_cur_state->f_texgen_eye_plane[m_p_cur_state->n_active_tex_unit][0], p_param, 4 * sizeof(GLfloat))) {
			memcpy(m_p_cur_state->f_texgen_eye_plane[m_p_cur_state->n_active_tex_unit][0], p_param, 4 * sizeof(GLfloat));
			glTexGenfv(GL_S, GL_EYE_PLANE, p_param);
		}
	} else {
		_ASSERTE(0);
	}
}

void __fastcall CGLState::TexGen_T_fv(int n_name, GLfloat *p_param)
{
	__FuncGuard("CGLState::TexGen_T_fv");

	if(n_name == GL_OBJECT_PLANE) {
		if(memcmp(m_p_cur_state->f_texgen_object_plane[m_p_cur_state->n_active_tex_unit][1], p_param, 4 * sizeof(GLfloat))) {
			memcpy(m_p_cur_state->f_texgen_object_plane[m_p_cur_state->n_active_tex_unit][1], p_param, 4 * sizeof(GLfloat));
			glTexGenfv(GL_T, GL_OBJECT_PLANE, p_param);
		}
	} else if(n_name == GL_EYE_PLANE) {
		if(memcmp(m_p_cur_state->f_texgen_eye_plane[m_p_cur_state->n_active_tex_unit][1], p_param, 4 * sizeof(GLfloat))) {
			memcpy(m_p_cur_state->f_texgen_eye_plane[m_p_cur_state->n_active_tex_unit][1], p_param, 4 * sizeof(GLfloat));
			glTexGenfv(GL_T, GL_EYE_PLANE, p_param);
		}
	} else {
		_ASSERTE(0);
	}
}

void __fastcall CGLState::TexGen_R_fv(int n_name, GLfloat *p_param)
{
	__FuncGuard("CGLState::TexGen_R_fv");

	if(n_name == GL_OBJECT_PLANE) {
		if(memcmp(m_p_cur_state->f_texgen_object_plane[m_p_cur_state->n_active_tex_unit][2], p_param, 4 * sizeof(GLfloat))) {
			memcpy(m_p_cur_state->f_texgen_object_plane[m_p_cur_state->n_active_tex_unit][2], p_param, 4 * sizeof(GLfloat));
			glTexGenfv(GL_R, GL_OBJECT_PLANE, p_param);
		}
	} else if(n_name == GL_EYE_PLANE) {
		if(memcmp(m_p_cur_state->f_texgen_eye_plane[m_p_cur_state->n_active_tex_unit][2], p_param, 4 * sizeof(GLfloat))) {
			memcpy(m_p_cur_state->f_texgen_eye_plane[m_p_cur_state->n_active_tex_unit][2], p_param, 4 * sizeof(GLfloat));
			glTexGenfv(GL_R, GL_EYE_PLANE, p_param);
		}
	} else {
		_ASSERTE(0);
	}
}

void __fastcall CGLState::TexGen_Q_fv(int n_name, GLfloat *p_param)
{
	__FuncGuard("CGLState::TexGen_Q_fv");

	if(n_name == GL_OBJECT_PLANE) {
		if(memcmp(m_p_cur_state->f_texgen_object_plane[m_p_cur_state->n_active_tex_unit][3], p_param, 4 * sizeof(GLfloat))) {
			memcpy(m_p_cur_state->f_texgen_object_plane[m_p_cur_state->n_active_tex_unit][3], p_param, 4 * sizeof(GLfloat));
			glTexGenfv(GL_Q, GL_OBJECT_PLANE, p_param);
		}
	} else if(n_name == GL_EYE_PLANE) {
		if(memcmp(m_p_cur_state->f_texgen_eye_plane[m_p_cur_state->n_active_tex_unit][3], p_param, 4 * sizeof(GLfloat))) {
			memcpy(m_p_cur_state->f_texgen_eye_plane[m_p_cur_state->n_active_tex_unit][3], p_param, 4 * sizeof(GLfloat));
			glTexGenfv(GL_Q, GL_EYE_PLANE, p_param);
		}
	} else {
		_ASSERTE(0);
	}
}

void __fastcall CGLState::TexGen_ST_fv(int n_name, GLfloat *p_param)
{
	__FuncGuard("CGLState::TexGen_ST_fv");

	if(n_name == GL_OBJECT_PLANE) {
		if(memcmp(m_p_cur_state->f_texgen_object_plane[m_p_cur_state->n_active_tex_unit][0], p_param, 4 * sizeof(GLfloat))) {
			memcpy(m_p_cur_state->f_texgen_object_plane[m_p_cur_state->n_active_tex_unit][0], p_param, 4 * sizeof(GLfloat));
			glTexGenfv(GL_S, GL_OBJECT_PLANE, p_param);
		}
		if(memcmp(m_p_cur_state->f_texgen_object_plane[m_p_cur_state->n_active_tex_unit][1], p_param, 4 * sizeof(GLfloat))) {
			memcpy(m_p_cur_state->f_texgen_object_plane[m_p_cur_state->n_active_tex_unit][1], p_param, 4 * sizeof(GLfloat));
			glTexGenfv(GL_T, GL_OBJECT_PLANE, p_param);
		}
	} else if(n_name == GL_EYE_PLANE) {
		if(memcmp(m_p_cur_state->f_texgen_eye_plane[m_p_cur_state->n_active_tex_unit][0], p_param, 4 * sizeof(GLfloat))) {
			memcpy(m_p_cur_state->f_texgen_eye_plane[m_p_cur_state->n_active_tex_unit][0], p_param, 4 * sizeof(GLfloat));
			glTexGenfv(GL_S, GL_EYE_PLANE, p_param);
		}
		if(memcmp(m_p_cur_state->f_texgen_eye_plane[m_p_cur_state->n_active_tex_unit][1], p_param, 4 * sizeof(GLfloat))) {
			memcpy(m_p_cur_state->f_texgen_eye_plane[m_p_cur_state->n_active_tex_unit][1], p_param, 4 * sizeof(GLfloat));
			glTexGenfv(GL_T, GL_EYE_PLANE, p_param);
		}
	} else {
		_ASSERTE(0);
	}
}

void __fastcall CGLState::TexGen_STR_fv(int n_name, GLfloat *p_param)
{
	__FuncGuard("CGLState::TexGen_STR_fv");

	if(n_name == GL_OBJECT_PLANE) {
		if(memcmp(m_p_cur_state->f_texgen_object_plane[m_p_cur_state->n_active_tex_unit][0], p_param, 4 * sizeof(GLfloat))) {
			memcpy(m_p_cur_state->f_texgen_object_plane[m_p_cur_state->n_active_tex_unit][0], p_param, 4 * sizeof(GLfloat));
			glTexGenfv(GL_S, GL_OBJECT_PLANE, p_param);
		}
		if(memcmp(m_p_cur_state->f_texgen_object_plane[m_p_cur_state->n_active_tex_unit][1], p_param, 4 * sizeof(GLfloat))) {
			memcpy(m_p_cur_state->f_texgen_object_plane[m_p_cur_state->n_active_tex_unit][1], p_param, 4 * sizeof(GLfloat));
			glTexGenfv(GL_T, GL_OBJECT_PLANE, p_param);
		}
		if(memcmp(m_p_cur_state->f_texgen_object_plane[m_p_cur_state->n_active_tex_unit][2], p_param, 4 * sizeof(GLfloat))) {
			memcpy(m_p_cur_state->f_texgen_object_plane[m_p_cur_state->n_active_tex_unit][2], p_param, 4 * sizeof(GLfloat));
			glTexGenfv(GL_R, GL_OBJECT_PLANE, p_param);
		}
	} else if(n_name == GL_EYE_PLANE) {
		if(memcmp(m_p_cur_state->f_texgen_eye_plane[m_p_cur_state->n_active_tex_unit][0], p_param, 4 * sizeof(GLfloat))) {
			memcpy(m_p_cur_state->f_texgen_eye_plane[m_p_cur_state->n_active_tex_unit][0], p_param, 4 * sizeof(GLfloat));
			glTexGenfv(GL_S, GL_EYE_PLANE, p_param);
		}
		if(memcmp(m_p_cur_state->f_texgen_eye_plane[m_p_cur_state->n_active_tex_unit][1], p_param, 4 * sizeof(GLfloat))) {
			memcpy(m_p_cur_state->f_texgen_eye_plane[m_p_cur_state->n_active_tex_unit][1], p_param, 4 * sizeof(GLfloat));
			glTexGenfv(GL_T, GL_EYE_PLANE, p_param);
		}
		if(memcmp(m_p_cur_state->f_texgen_eye_plane[m_p_cur_state->n_active_tex_unit][2], p_param, 4 * sizeof(GLfloat))) {
			memcpy(m_p_cur_state->f_texgen_eye_plane[m_p_cur_state->n_active_tex_unit][2], p_param, 4 * sizeof(GLfloat));
			glTexGenfv(GL_R, GL_EYE_PLANE, p_param);
		}
	} else {
		_ASSERTE(0);
	}
}

void __fastcall CGLState::TexGenfv(int n_coord, int n_name, GLfloat *p_param)
{
	__FuncGuard("CGLState::TexGenfv");

	if(n_name == GL_OBJECT_PLANE) {
		if(memcmp(m_p_cur_state->f_texgen_object_plane[m_p_cur_state->n_active_tex_unit][n_coord], p_param, 4 * sizeof(GLfloat))) {
			memcpy(m_p_cur_state->f_texgen_object_plane[m_p_cur_state->n_active_tex_unit][n_coord], p_param, 4 * sizeof(GLfloat));
			glTexGenfv(GL_S + n_coord, GL_OBJECT_PLANE, p_param);
		}
	} else if(n_name == GL_EYE_PLANE) {
		if(memcmp(m_p_cur_state->f_texgen_eye_plane[m_p_cur_state->n_active_tex_unit][n_coord], p_param, 4 * sizeof(GLfloat))) {
			memcpy(m_p_cur_state->f_texgen_eye_plane[m_p_cur_state->n_active_tex_unit][n_coord], p_param, 4 * sizeof(GLfloat));
			glTexGenfv(GL_S + n_coord, GL_EYE_PLANE, p_param);
		}
	} else {
		_ASSERTE(0);
	}
}

void __fastcall CGLState::TexGen_STRQ_Mode(/*int n_name, */GLenum n_param)
{
	__FuncGuard("CGLState::TexGen_STRQ_Mode");

	if(m_p_cur_state->n_texgen_mode[m_p_cur_state->n_active_tex_unit][0] != n_param)
		glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, m_p_cur_state->n_texgen_mode[m_p_cur_state->n_active_tex_unit][0] = n_param);
	if(m_p_cur_state->n_texgen_mode[m_p_cur_state->n_active_tex_unit][1] != n_param)
		glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, m_p_cur_state->n_texgen_mode[m_p_cur_state->n_active_tex_unit][1] = n_param);
	if(m_p_cur_state->n_texgen_mode[m_p_cur_state->n_active_tex_unit][2] != n_param)
		glTexGeni(GL_R, GL_TEXTURE_GEN_MODE, m_p_cur_state->n_texgen_mode[m_p_cur_state->n_active_tex_unit][2] = n_param);
	if(m_p_cur_state->n_texgen_mode[m_p_cur_state->n_active_tex_unit][3] != n_param)
		glTexGeni(GL_Q, GL_TEXTURE_GEN_MODE, m_p_cur_state->n_texgen_mode[m_p_cur_state->n_active_tex_unit][3] = n_param);
}

void __fastcall CGLState::TexGen_STR_Mode(/*int n_name, */GLenum n_param)
{
	__FuncGuard("CGLState::TexGen_STR_Mode");

	if(m_p_cur_state->n_texgen_mode[m_p_cur_state->n_active_tex_unit][0] != n_param)
		glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, m_p_cur_state->n_texgen_mode[m_p_cur_state->n_active_tex_unit][0] = n_param);
	if(m_p_cur_state->n_texgen_mode[m_p_cur_state->n_active_tex_unit][1] != n_param)
		glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, m_p_cur_state->n_texgen_mode[m_p_cur_state->n_active_tex_unit][1] = n_param);
	if(m_p_cur_state->n_texgen_mode[m_p_cur_state->n_active_tex_unit][2] != n_param)
		glTexGeni(GL_R, GL_TEXTURE_GEN_MODE, m_p_cur_state->n_texgen_mode[m_p_cur_state->n_active_tex_unit][2] = n_param);
}

void __fastcall CGLState::TexGen_S_Mode(/*int n_name, */GLenum n_param)
{
	__FuncGuard("CGLState::TexGen_S_Mode");

	if(m_p_cur_state->n_texgen_mode[m_p_cur_state->n_active_tex_unit][0] != n_param)
		glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, m_p_cur_state->n_texgen_mode[m_p_cur_state->n_active_tex_unit][0] = n_param);
}

void __fastcall CGLState::TexGen_T_Mode(/*int n_name, */GLenum n_param)
{
	__FuncGuard("CGLState::TexGen_T_Mode");

	if(m_p_cur_state->n_texgen_mode[m_p_cur_state->n_active_tex_unit][1] != n_param)
		glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, m_p_cur_state->n_texgen_mode[m_p_cur_state->n_active_tex_unit][1] = n_param);
}

void __fastcall CGLState::TexGen_R_Mode(/*int n_name, */GLenum n_param)
{
	__FuncGuard("CGLState::TexGen_R_Mode");

	if(m_p_cur_state->n_texgen_mode[m_p_cur_state->n_active_tex_unit][2] != n_param)
		glTexGeni(GL_R, GL_TEXTURE_GEN_MODE, m_p_cur_state->n_texgen_mode[m_p_cur_state->n_active_tex_unit][2] = n_param);
}

void __fastcall CGLState::TexGen_Q_Mode(/*int n_name, */GLenum n_param)
{
	__FuncGuard("CGLState::TexGen_Q_Mode");

	if(m_p_cur_state->n_texgen_mode[m_p_cur_state->n_active_tex_unit][3] != n_param)
		glTexGeni(GL_Q, GL_TEXTURE_GEN_MODE, m_p_cur_state->n_texgen_mode[m_p_cur_state->n_active_tex_unit][3] = n_param);
}

void __fastcall CGLState::TexGen_ST_Mode(/*int n_name, */GLenum n_param)
{
	__FuncGuard("CGLState::TexGen_ST_Mode");

	if(m_p_cur_state->n_texgen_mode[m_p_cur_state->n_active_tex_unit][0] != n_param)
		glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, m_p_cur_state->n_texgen_mode[m_p_cur_state->n_active_tex_unit][0] = n_param);
	if(m_p_cur_state->n_texgen_mode[m_p_cur_state->n_active_tex_unit][1] != n_param)
		glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, m_p_cur_state->n_texgen_mode[m_p_cur_state->n_active_tex_unit][1] = n_param);
}

void __fastcall CGLState::TexGenMode(int n_coord, GLenum n_param)
{
	__FuncGuard("CGLState::TexGenMode");

	if(m_p_cur_state->n_texgen_mode[m_p_cur_state->n_active_tex_unit][n_coord] != n_param)
		glTexGeni(GL_S + n_coord, GL_TEXTURE_GEN_MODE, m_p_cur_state->n_texgen_mode[m_p_cur_state->n_active_tex_unit][n_coord] = n_param);
}

void __fastcall CGLState::ColorMask(bool b_r, bool b_g, bool b_b, bool b_a)
{
	__FuncGuard("CGLState::ColorMask");

	if(m_p_cur_state->b_color_write[0] != b_r ||
	   m_p_cur_state->b_color_write[1] != b_g ||
	   m_p_cur_state->b_color_write[2] != b_b ||
	   m_p_cur_state->b_color_write[3] != b_a)
		glColorMask(m_p_cur_state->b_color_write[0] = b_r,
					m_p_cur_state->b_color_write[1] = b_g,
					m_p_cur_state->b_color_write[2] = b_b,
					m_p_cur_state->b_color_write[3] = b_a);
}

void __fastcall CGLState::DepthMask(bool b_d)
{
	__FuncGuard("CGLState::DepthMask");

	if(m_p_cur_state->b_depth_write != b_d)
		glDepthMask(m_p_cur_state->b_depth_write = b_d);
}

void __fastcall CGLState::StencilMask(GLuint n_mask)
{
	__FuncGuard("CGLState::StencilMask");

	if(m_p_cur_state->n_stencil_write[0] != n_mask ||
	   m_p_cur_state->n_stencil_write[1] != n_mask) {
		if(m_p_cur_state->b_separate_gl2_stencil) {
			glStencilMaskSeparate(GL_FRONT_AND_BACK, m_p_cur_state->n_stencil_write[0] =
				m_p_cur_state->n_stencil_write[1] = n_mask);
		} else
			glStencilMask(m_p_cur_state->n_stencil_write[1] = n_mask);
	}
}

void __fastcall CGLState::StencilMask(bool b_front, GLuint n_mask)
{
	__FuncGuard("CGLState::StencilMask");

	_ASSERTE(b_front == !!b_front);
	if(m_p_cur_state->n_stencil_write[b_front] != n_mask) {
		_ASSERTE(m_p_cur_state->b_separate_gl2_stencil);
		glStencilMaskSeparate((b_front)? GL_FRONT : GL_BACK,
			m_p_cur_state->n_stencil_write[b_front] = n_mask);
	}
}

void __fastcall CGLState::Enable(GLenum n_value)
{
	__FuncGuard("CGLState::Enable");

	switch(n_value) {
	case GL_ALPHA_TEST:
		EnableAlphaTest();
		return;
	case GL_DEPTH_TEST:
		EnableDepthTest();
		return;
	case GL_CULL_FACE:
		EnableCullFace();
		return;
	case GL_TEXTURE_1D:
		EnableTexture1D();
		return;
	case GL_TEXTURE_2D:
		EnableTexture2D();
		return;
	case GL_TEXTURE_3D:
		EnableTexture3D();
		return;
	case GL_TEXTURE_CUBE_MAP:
		EnableTextureCubeMap();
		return;
	case GL_BLEND:
		EnableBlend();
		return;
	case GL_LIGHTING:
		EnableLighting();
		return;
	/*case GL_LIGHT0:
	case GL_LIGHT1:
	case GL_LIGHT2:
	case GL_LIGHT3:
	case GL_LIGHT4:
	case GL_LIGHT5:
	case GL_LIGHT6:
	case GL_LIGHT7:
		EnableLight(n_value - GL_LIGHT0);
		return;*/ // below
	case GL_FOG:
		EnableFog();
		return;
	case GL_TEXTURE_GEN_S:
		Enable_S_TexGen();
		return;
	case GL_TEXTURE_GEN_T:
		Enable_T_TexGen();
		return;
	case GL_TEXTURE_GEN_R:
		Enable_R_TexGen();
		return;
	case GL_TEXTURE_GEN_Q:
		Enable_Q_TexGen();
		return;
	case GL_POLYGON_OFFSET_LINE:
		EnablePolygonOffset_Line();
		return;
	case GL_POLYGON_OFFSET_FILL:
		EnablePolygonOffset_Fill();
		return;
	case GL_POLYGON_OFFSET_POINT:
		EnablePolygonOffset_Point();
		return;

	default:
		if(n_value >= GL_CLIP_PLANE0 && n_value < GL_CLIP_PLANE0 + 32)
			EnableClipPlane(n_value - GL_CLIP_PLANE0);
		else if(n_value >= GL_LIGHT0 && n_value < GL_LIGHT0 + 32)
			EnableLight(n_value - GL_LIGHT0);
		else
			glEnable(n_value);
	}
}

void __fastcall CGLState::Disable(GLenum n_value)
{
	__FuncGuard("CGLState::Disable");

	switch(n_value) {
	case GL_ALPHA_TEST:
		DisableAlphaTest();
		return;
	case GL_DEPTH_TEST:
		DisableDepthTest();
		return;
	case GL_CULL_FACE:
		DisableCullFace();
		return;
	case GL_TEXTURE_1D:
		DisableTexture1D();
		return;
	case GL_TEXTURE_2D:
		DisableTexture2D();
		return;
	case GL_TEXTURE_3D:
		DisableTexture3D();
		return;
	case GL_TEXTURE_CUBE_MAP:
		DisableTextureCubeMap();
		return;
	case GL_BLEND:
		DisableBlend();
		return;
	case GL_LIGHTING:
		DisableLighting();
		return;
	/*case GL_LIGHT0:
	case GL_LIGHT1:
	case GL_LIGHT2:
	case GL_LIGHT3:
	case GL_LIGHT4:
	case GL_LIGHT5:
	case GL_LIGHT6:
	case GL_LIGHT7:
		DisableLight(n_value - GL_LIGHT0);
		return;*/ // below
	case GL_FOG:
		DisableFog();
		return;
	case GL_TEXTURE_GEN_S:
		Disable_S_TexGen();
		return;
	case GL_TEXTURE_GEN_T:
		Disable_T_TexGen();
		return;
	case GL_TEXTURE_GEN_R:
		Disable_R_TexGen();
		return;
	case GL_TEXTURE_GEN_Q:
		Disable_Q_TexGen();
		return;
	case GL_POLYGON_OFFSET_LINE:
		DisablePolygonOffset_Line();
		return;
	case GL_POLYGON_OFFSET_FILL:
		DisablePolygonOffset_Fill();
		return;
	case GL_POLYGON_OFFSET_POINT:
		DisablePolygonOffset_Point();
		return;

	default:
		if(n_value >= GL_CLIP_PLANE0 && n_value < GL_CLIP_PLANE0 + 32)
			DisableClipPlane(n_value - GL_CLIP_PLANE0);
		else if(n_value >= GL_LIGHT0 && n_value < GL_LIGHT0 + 32)
			DisableLight(n_value - GL_LIGHT0);
		else
			glDisable(n_value);
	}
}

void __fastcall CGLState::EnablePolygonOffset(GLenum n_mode)
{
	__FuncGuard("CGLState::EnablePolygonOffset");

	int n = 0;
	switch(n_mode) {
	case GL_POLYGON_OFFSET_FILL:
		n = 0;
		break;
	case GL_POLYGON_OFFSET_LINE:
		n = 1;
		break;
	case GL_POLYGON_OFFSET_POINT:
		n = 2;
		break;
	}

	if(m_p_cur_state->b_polygon_offset_support && !m_p_cur_state->b_polygon_offset[n]) {
		glEnable(n_mode);
		m_p_cur_state->b_polygon_offset[n] = true;
	}
}

void __fastcall CGLState::DisablePolygonOffset(GLenum n_mode)
{
	__FuncGuard("CGLState::DisablePolygonOffset");

	int n = 0;
	switch(n_mode) {
	case GL_POLYGON_OFFSET_FILL:
		n = 0;
		break;
	case GL_POLYGON_OFFSET_LINE:
		n = 1;
		break;
	case GL_POLYGON_OFFSET_POINT:
		n = 2;
		break;
	}

	if(m_p_cur_state->b_polygon_offset_support && m_p_cur_state->b_polygon_offset[n]) {
		glDisable(n_mode);
		m_p_cur_state->b_polygon_offset[n] = false;
	}
}

void __fastcall CGLState::EnablePolygonOffset_Fill()
{
	__FuncGuard("CGLState::EnablePolygonOffset_Fill");

	if(m_p_cur_state->b_polygon_offset_support && !m_p_cur_state->b_polygon_offset[0]) {
		glEnable(GL_POLYGON_OFFSET_FILL);
		m_p_cur_state->b_polygon_offset[0] = true;
	}
}

void __fastcall CGLState::DisablePolygonOffset_Fill()
{
	__FuncGuard("CGLState::DisablePolygonOffset_Fill");

	if(m_p_cur_state->b_polygon_offset_support && m_p_cur_state->b_polygon_offset[0]) {
		glDisable(GL_POLYGON_OFFSET_FILL);
		m_p_cur_state->b_polygon_offset[0] = false;
	}
}

void __fastcall CGLState::EnablePolygonOffset_Line()
{
	__FuncGuard("CGLState::EnablePolygonOffset_Line");

	if(m_p_cur_state->b_polygon_offset_support && !m_p_cur_state->b_polygon_offset[1]) {
		glEnable(GL_POLYGON_OFFSET_LINE);
		m_p_cur_state->b_polygon_offset[1] = true;
	}
}

void __fastcall CGLState::DisablePolygonOffset_Line()
{
	__FuncGuard("CGLState::DisablePolygonOffset_Line");

	if(m_p_cur_state->b_polygon_offset_support && m_p_cur_state->b_polygon_offset[1]) {
		glDisable(GL_POLYGON_OFFSET_LINE);
		m_p_cur_state->b_polygon_offset[1] = false;
	}
}

void __fastcall CGLState::EnablePolygonOffset_Point()
{
	__FuncGuard("CGLState::EnablePolygonOffset_Point");

	if(m_p_cur_state->b_polygon_offset_support && !m_p_cur_state->b_polygon_offset[2]) {
		glEnable(GL_POLYGON_OFFSET_POINT);
		m_p_cur_state->b_polygon_offset[2] = true;
	}
}

void __fastcall CGLState::DisablePolygonOffset_Point()
{
	__FuncGuard("CGLState::DisablePolygonOffset_Point");

	if(m_p_cur_state->b_polygon_offset_support && m_p_cur_state->b_polygon_offset[2]) {
		glDisable(GL_POLYGON_OFFSET_POINT);
		m_p_cur_state->b_polygon_offset[2] = false;
	}
}

void __fastcall CGLState::PolygonOffset(GLfloat f_factor, GLfloat f_bias)
{
	__FuncGuard("CGLState::PolygonOffset");

	if(m_p_cur_state->b_polygon_offset_support) {
		if(m_p_cur_state->f_polygon_offset_factor != f_factor ||
		   m_p_cur_state->f_polygon_offset_bias != f_bias) {
			glPolygonOffset(m_p_cur_state->f_polygon_offset_factor = f_factor,
							   m_p_cur_state->f_polygon_offset_bias = f_bias);
		}
	}
}

void __fastcall CGLState::EnableVertexProgram()
{
	__FuncGuard("CGLState::EnableVertexProgram");

	if(m_p_cur_state->b_vertex_program_support && !m_p_cur_state->b_vertex_program_enabled) {
		glEnable(GL_VERTEX_PROGRAM_ARB);
		m_p_cur_state->b_vertex_program_enabled = true;
	}
}

void __fastcall CGLState::EnableFragmentProgram()
{
	__FuncGuard("CGLState::EnableFragmentProgram");

	if(m_p_cur_state->b_fragment_program_support && !m_p_cur_state->b_fragment_program_enabled) {
		glEnable(GL_FRAGMENT_PROGRAM_ARB);
		m_p_cur_state->b_fragment_program_enabled = true;
	}
}

void __fastcall CGLState::DisableVertexProgram()
{
	__FuncGuard("CGLState::DisableVertexProgram");

	if(m_p_cur_state->b_vertex_program_support && m_p_cur_state->b_vertex_program_enabled) {
		glDisable(GL_VERTEX_PROGRAM_ARB);
		m_p_cur_state->b_vertex_program_enabled = false;
	}
}

void __fastcall CGLState::DisableFragmentProgram()
{
	__FuncGuard("CGLState::DisableFragmentProgram");

	if(m_p_cur_state->b_fragment_program_support && m_p_cur_state->b_fragment_program_enabled) {
		glDisable(GL_FRAGMENT_PROGRAM_ARB);
		m_p_cur_state->b_fragment_program_enabled = false;
	}
}

void __fastcall CGLState::BindProgramObject(GLhandleARB h_shader)
{
	__FuncGuard("CGLState::BindProgramObject");

	if(m_p_cur_state->b_shader_support && m_p_cur_state->n_program_object != h_shader) {
		glUseProgramObjectARB(m_p_cur_state->n_program_object = h_shader);
	}
}

void __fastcall CGLState::BindProgramObject_GL2(GLuint n_shader)
{
	__FuncGuard("CGLState::BindProgramObject_GL2");

	if(m_p_cur_state->b_shader_gl2_support && m_p_cur_state->n_program_object != n_shader)
		glUseProgram(m_p_cur_state->n_program_object = n_shader);
}

void __fastcall CGLState::BindVertexProgram(GLuint n_program)
{
	__FuncGuard("CGLState::BindVertexProgram");

	if(m_p_cur_state->b_vertex_program_support && m_p_cur_state->n_vertex_program != n_program)
		glBindProgramARB(GL_VERTEX_PROGRAM_ARB, m_p_cur_state->n_vertex_program = n_program);
}

void __fastcall CGLState::BindFragmentProgram(GLuint n_program)
{
	__FuncGuard("CGLState::BindFragmentProgram");

	if(m_p_cur_state->b_vertex_program_support && m_p_cur_state->n_fragment_program != n_program)
		glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB, m_p_cur_state->n_fragment_program = n_program);
}

void __fastcall CGLState::LightModelAmbient(const GLfloat *p_ambient4)
{
	__FuncGuard("CGLState::LightModelAmbient");

	if(memcmp(m_p_cur_state->f_model_ambient, p_ambient4, 4 * sizeof(GLfloat))) {
		memcpy(m_p_cur_state->f_model_ambient, p_ambient4, 4 * sizeof(GLfloat));
		glLightModelfv(GL_LIGHT_MODEL_AMBIENT, p_ambient4);
	}
}

void __fastcall CGLState::LightModelViewer(int b_local)
{
	__FuncGuard("CGLState::LightModelViewer");

	if(m_p_cur_state->b_model_local_viewer != b_local)
		glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, m_p_cur_state->b_model_local_viewer = b_local);
}

void __fastcall CGLState::LightModelTwoSide(int b_two_side)
{
	__FuncGuard("CGLState::LightModelTwoSide");

	if(m_p_cur_state->b_model_two_side != b_two_side)
		glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, m_p_cur_state->b_model_two_side = b_two_side);
}

void __fastcall CGLState::LightModelColorControl(GLenum n_mode)
{
	__FuncGuard("CGLState::LightModelColorControl");

	if(m_p_cur_state->n_model_color_control != n_mode) {
		glLightModeli(GL_LIGHT_MODEL_COLOR_CONTROL,
			m_p_cur_state->n_model_color_control = n_mode);
	}
}

void __fastcall CGLState::MaterialAmbient(const GLfloat *p_ambient4)
{
	__FuncGuard("CGLState::MaterialAmbient");

	if(memcmp(m_p_cur_state->f_material_ambient[0], p_ambient4, 4 * sizeof(GLfloat)) ||
	   memcmp(m_p_cur_state->f_material_ambient[1], p_ambient4, 4 * sizeof(GLfloat))) {
		memcpy(m_p_cur_state->f_material_ambient[0], p_ambient4, 4 * sizeof(GLfloat));
		memcpy(m_p_cur_state->f_material_ambient[1], p_ambient4, 4 * sizeof(GLfloat));
		glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, p_ambient4);
	}
}

void __fastcall CGLState::MaterialAmbient(int b_front, const GLfloat *p_ambient4)
{
	__FuncGuard("CGLState::MaterialAmbient");

	_ASSERTE(b_front == !!b_front);
	if(memcmp(m_p_cur_state->f_material_ambient[b_front], p_ambient4, 4 * sizeof(GLfloat))) {
		memcpy(m_p_cur_state->f_material_ambient[b_front], p_ambient4, 4 * sizeof(GLfloat));
		glMaterialfv((b_front)? GL_FRONT : GL_BACK, GL_AMBIENT, p_ambient4);
	}
}

void __fastcall CGLState::MaterialDiffuse(const GLfloat *p_diffuse4)
{
	__FuncGuard("CGLState::MaterialDiffuse");

	if(memcmp(m_p_cur_state->f_material_diffuse[0], p_diffuse4, 4 * sizeof(GLfloat)) ||
	   memcmp(m_p_cur_state->f_material_diffuse[1], p_diffuse4, 4 * sizeof(GLfloat))) {
		memcpy(m_p_cur_state->f_material_diffuse[0], p_diffuse4, 4 * sizeof(GLfloat));
		memcpy(m_p_cur_state->f_material_diffuse[1], p_diffuse4, 4 * sizeof(GLfloat));
		glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, p_diffuse4);
	}
}

void __fastcall CGLState::MaterialDiffuse(int b_front, const GLfloat *p_diffuse4)
{
	__FuncGuard("CGLState::MaterialDiffuse");

	_ASSERTE(b_front == !!b_front);
	if(memcmp(m_p_cur_state->f_material_diffuse[b_front], p_diffuse4, 4 * sizeof(GLfloat))) {
		memcpy(m_p_cur_state->f_material_diffuse[b_front], p_diffuse4, 4 * sizeof(GLfloat));
		glMaterialfv((b_front)? GL_FRONT : GL_BACK, GL_DIFFUSE, p_diffuse4);
	}
}

void __fastcall CGLState::MaterialSpecular(const GLfloat *p_specular4)
{
	__FuncGuard("CGLState::MaterialSpecular");

	if(memcmp(m_p_cur_state->f_material_specular[0], p_specular4, 4 * sizeof(GLfloat)) ||
	   memcmp(m_p_cur_state->f_material_specular[1], p_specular4, 4 * sizeof(GLfloat))) {
		memcpy(m_p_cur_state->f_material_specular[0], p_specular4, 4 * sizeof(GLfloat));
		memcpy(m_p_cur_state->f_material_specular[1], p_specular4, 4 * sizeof(GLfloat));
		glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, p_specular4);
	}
}

void __fastcall CGLState::MaterialSpecular(int b_front, const GLfloat *p_specular4)
{
	__FuncGuard("CGLState::MaterialSpecular");

	_ASSERTE(b_front == !!b_front);
	if(memcmp(m_p_cur_state->f_material_specular[b_front], p_specular4, 4 * sizeof(GLfloat))) {
		memcpy(m_p_cur_state->f_material_specular[b_front], p_specular4, 4 * sizeof(GLfloat));
		glMaterialfv((b_front)? GL_FRONT : GL_BACK, GL_SPECULAR, p_specular4);
	}
}

void __fastcall CGLState::MaterialEmission(const GLfloat *p_emission4)
{
	__FuncGuard("CGLState::MaterialEmission");

	if(memcmp(m_p_cur_state->f_material_emission[0], p_emission4, 4 * sizeof(GLfloat)) ||
	   memcmp(m_p_cur_state->f_material_emission[1], p_emission4, 4 * sizeof(GLfloat))) {
		memcpy(m_p_cur_state->f_material_emission[0], p_emission4, 4 * sizeof(GLfloat));
		memcpy(m_p_cur_state->f_material_emission[1], p_emission4, 4 * sizeof(GLfloat));
		glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, p_emission4);
	}
}

void __fastcall CGLState::MaterialEmission(int b_front, const GLfloat *p_emission4)
{
	__FuncGuard("CGLState::MaterialEmission");

	_ASSERTE(b_front == !!b_front);
	if(memcmp(m_p_cur_state->f_material_emission[b_front], p_emission4, 4 * sizeof(GLfloat))) {
		memcpy(m_p_cur_state->f_material_emission[b_front], p_emission4, 4 * sizeof(GLfloat));
		glMaterialfv((b_front)? GL_FRONT : GL_BACK, GL_EMISSION, p_emission4);
	}
}

void __fastcall CGLState::MaterialShininess(GLfloat f_shininess)
{
	__FuncGuard("CGLState::MaterialShininess");

	if(m_p_cur_state->f_material_shininess[0] != f_shininess ||
	   m_p_cur_state->f_material_shininess[1] != f_shininess) {
		glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS,
			m_p_cur_state->f_material_shininess[0] =
			m_p_cur_state->f_material_shininess[1] = f_shininess);
	}
}

void __fastcall CGLState::MaterialShininess(int b_front, GLfloat f_shininess)
{
	__FuncGuard("CGLState::MaterialShininess");

	_ASSERTE(b_front == !!b_front);
	if(m_p_cur_state->f_material_shininess[b_front] != f_shininess) {
		glMaterialf((b_front)? GL_FRONT : GL_BACK, GL_SHININESS,
			m_p_cur_state->f_material_shininess[b_front] = f_shininess);
	}
}

void __fastcall CGLState::EnableClipPlane(int n_plane)
{
	__FuncGuard("CGLState::EnableClipPlane");

	_ASSERTE(n_plane >= 0 && n_plane < m_p_cur_state->n_max_clip_planes);
	if(!m_p_cur_state->b_clip_plane[n_plane]) {
		m_p_cur_state->b_clip_plane[n_plane] = true;
		glEnable(GL_CLIP_PLANE0 + n_plane);
	}
}

void __fastcall CGLState::DisableClipPlane(int n_plane)
{
	__FuncGuard("CGLState::DisableClipPlane");

	_ASSERTE(n_plane >= 0 && n_plane < m_p_cur_state->n_max_clip_planes);
	if(m_p_cur_state->b_clip_plane[n_plane]) {
		m_p_cur_state->b_clip_plane[n_plane] = false;
		glDisable(GL_CLIP_PLANE0 + n_plane);
	}
}

void __fastcall CGLState::DisableAllClipPlanes()
{
	__FuncGuard("CGLState::DisableAllClipPlanes");

	for(int i = 0; i < m_p_cur_state->n_max_clip_planes; i ++) {
		if(m_p_cur_state->b_clip_plane[i]) {
			m_p_cur_state->b_clip_plane[i] = false;
			glDisable(GL_CLIP_PLANE0 + i);
		}
	}
}

void __fastcall CGLState::ClipPlane(int n_plane, double f_eqn0, double f_eqn1, double f_eqn2, double f_eqn3)
{
	__FuncGuard("CGLState::ClipPlane");

	_ASSERTE(n_plane >= 0 && n_plane < m_p_cur_state->n_max_clip_planes);
	if(m_p_cur_state->f_clip_plane[n_plane][0] != f_eqn0 ||
	   m_p_cur_state->f_clip_plane[n_plane][1] != f_eqn1 ||
	   m_p_cur_state->f_clip_plane[n_plane][2] != f_eqn2 ||
	   m_p_cur_state->f_clip_plane[n_plane][3] != f_eqn3) {
		m_p_cur_state->f_clip_plane[n_plane][0] = f_eqn0;
		m_p_cur_state->f_clip_plane[n_plane][1] = f_eqn1;
		m_p_cur_state->f_clip_plane[n_plane][2] = f_eqn2;
		m_p_cur_state->f_clip_plane[n_plane][3] = f_eqn3;
		glClipPlane(GL_CLIP_PLANE0 + n_plane, m_p_cur_state->f_clip_plane[n_plane]);
	}
}

/*
 *								=== ~CGLState ===
 */

/*
 *		-end-of-file-
 */
